from typing import List


class QueryValidation:
    """Helper object for validating subgraph queries."""

    @classmethod
    def validate_offer_query(
        cls,
        order_by: str,
        order_direction: str,
        first: int,
    ):
        """Validate the offer query."""
        error_messages = []

        error_messages.extend(
            cls.general_validations(
                order_by=order_by,
                allowed_order_by=["timestamp", "price"],
                order_direction=order_direction,
                first=first,
            )
        )

        if error_messages:
            raise ValueError("\n".join(error_messages))

    @classmethod
    def validate_trade_query(
        cls,
        order_by: str,
        order_direction: str,
        first: int,
    ):
        """Validate the trade query."""
        error_messages = []

        error_messages.extend(
            cls.general_validations(
                order_by=order_by,
                allowed_order_by=["timestamp"],
                order_direction=order_direction,
                first=first,
            )
        )

        if error_messages:
            raise ValueError("\n".join(error_messages))

    @staticmethod
    def general_validations(
        order_by: str,
        allowed_order_by: List[str],
        order_direction: str,
        first: int,
    ) -> List[str]:
        """General validations helper."""
        error_messages = []

        # check the order_by parameter
        if order_by.lower() not in allowed_order_by:
            error_messages.append(
                f"Invalid order_by, must be on of '{allowed_order_by}' (default: timestamp)"
            )

        # check the order_direction parameter
        if order_direction.lower() not in ("asc", "desc"):
            error_messages.append(
                "Invalid order_direction, must be 'asc' or 'desc' (default: desc)"
            )

        # check the first parameter
        if first < 1:
            error_messages.append(
                "Invalid first, must be greater than 0 (default: 1000)"
            )
        if not isinstance(first, int):
            error_messages.append("Invalid first, must be an integer (default: 1000)")

        return error_messages



from _decimal import Decimal
from enum import Enum
from typing import Optional, List, Any, Union

from eth_typing import BlockNumber, ChecksumAddress
from hexbytes import HexBytes
from web3.types import Wei, TxReceipt, EventData

from rubi.contracts.contract_types import BaseEvent


class TransactionStatus(Enum):
    """Enum representing the status of a transaction."""

    SUCCESS = "SUCCESS"
    FAILURE = "FAILURE"

    def __eq__(self, other):
        return self.name is other.name and self.value == other.value

    @classmethod
    def from_int(cls, status: int) -> "TransactionStatus":
        """
        Transform the TxReceipt status into an Enum
        """
        if status == 0:
            return cls.FAILURE
        else:
            return cls.SUCCESS


class TransactionReceipt:
    """Transaction receipt object"""

    def __init__(
        self,
        block_number: BlockNumber,
        contract_address: Optional[ChecksumAddress],
        effective_gas_price: Wei,
        gas_used: int,
        from_address: ChecksumAddress,
        to_address: ChecksumAddress,
        status: int,
        transaction_hash: HexBytes,
        transaction_index: int,
        l1_fee: Optional[int] = None,
        l1_gas_price: Optional[int] = None,
        l1_gas_used: Optional[int] = None,
        l1_fee_scalar: Optional[Decimal] = None,
        raw_events: Optional[List[BaseEvent]] = None,
    ):
        # Block number
        self.block_number = block_number

        # Addresses
        self.from_address = from_address
        self.to_address = to_address
        self.created_contract_at_address = contract_address

        # Transaction cost
        self.gas_used = gas_used
        self.transaction_cost_in_eth = self._calc_gas_cost_in_eth(
            effective_gas_price=effective_gas_price,
            gas_used=gas_used,
            l1_gas_price=l1_gas_price,
            l1_gas_used=l1_gas_used,
            l1_fee_scalar=l1_fee_scalar,
        )

        # Transaction details
        self.transaction_status = TransactionStatus.from_int(status=status)
        self.transaction_hash = transaction_hash.hex()
        self.position_in_block = transaction_index

        # Transaction Events
        self.raw_events = raw_events
        self.events = None

    @classmethod
    def from_tx_receipt(
        cls, tx_receipt: TxReceipt, raw_events: List[Union[BaseEvent, EventData]]
    ) -> "TransactionReceipt":
        """Initialize a TransactionReceipt

        :param tx_receipt: The transaction receipt Dict received from the node.
        :type tx_receipt: TxReceipt
        :param raw_events: The raw events decoded from the logs
        :type raw_events: List[Union[BaseEvent, EventData]]
        :return: A TransactionReceipt object
        :rtype: TransactionReceipt
        """
        return cls(
            block_number=tx_receipt["blockNumber"],
            contract_address=tx_receipt["contractAddress"],
            effective_gas_price=tx_receipt["effectiveGasPrice"],
            gas_used=tx_receipt["gasUsed"],
            from_address=tx_receipt["from"],
            to_address=tx_receipt["to"],
            status=tx_receipt["status"],
            transaction_hash=tx_receipt["transactionHash"],
            transaction_index=tx_receipt["transactionIndex"],
            l1_fee=None
            if tx_receipt.get("l1Fee") is None
            else int(tx_receipt.get("l1Fee"), 16),
            l1_gas_price=None
            if tx_receipt.get("l1GasPrice") is None
            else int(tx_receipt.get("l1GasPrice"), 16),
            l1_gas_used=None
            if tx_receipt.get("l1GasUsed") is None
            else int(tx_receipt.get("l1GasUsed"), 16),
            l1_fee_scalar=None
            if tx_receipt.get("l1FeeScalar") is None
            else Decimal(tx_receipt.get("l1FeeScalar")),
            raw_events=raw_events,
        )

    # TODO: Any is used to avoid circular dependencies, look at a restructure. These are OrderEvents | ApprovalEvents |
    #  TransferEvents.
    def set_events(self, events: List[Any]):
        """Set the events on the Transaction Receipt. Used on the client to set human-readable events"""
        self.events = events

    @staticmethod
    def _calc_gas_cost_in_eth(
        gas_used: int,
        effective_gas_price: Wei,
        l1_gas_price: Optional[int] = None,
        l1_gas_used: Optional[int] = None,
        l1_fee_scalar: Optional[Decimal] = None,
    ) -> Optional[Decimal]:
        """Calculate the cost of the transaction in eth.

        :param gas_used: The amount of gas used by the transaction.
        :type gas_used: int
        :param effective_gas_price: The effective L2 gas price of the transaction.
        :type effective_gas_price: Wei
        :param l1_gas_price: The L1 gas price.
        :type l1_gas_price: Optional[int]
        :param l1_gas_used: The L1 gas used.
        :type l1_gas_used: Optional[int]
        :param l1_fee_scalar: The L1 fee scalar.
        :type l1_fee_scalar: Optional[int]
        :return:
        """
        if not (l1_gas_price and l1_gas_used and l1_fee_scalar):
            return None

        return (
            Decimal(effective_gas_price * gas_used)
            + (l1_gas_price * l1_gas_used * l1_fee_scalar)
        ) / Decimal(10**18)

    def __repr__(self):
        items = (
            "{}={!r}".format(k, self.__dict__[k])
            for k in self.__dict__
            if k != "raw_events"
        )
        return "{}({})".format(type(self).__name__, ", ".join(items))



from .contract_types import *
from .base_contract import BaseContract
from .erc20 import ERC20
from .market import RubiconMarket
from .router import RubiconRouter
from .transaction_handler import TransactionHandler



from typing import Any, List

from eth_typing import ChecksumAddress


class SubgraphOffer:
    """Helper object for querying subgraph Offers"""

    def __init__(
        self,
        order_id: int,
        order_owner: ChecksumAddress,
        pay_gem: ChecksumAddress,
        pay_amt: int,
        paid_amt: int,
        buy_gem: ChecksumAddress,
        buy_amt: int,
        bought_amt: int,
        open: bool,
    ):
        self.order_id = order_id
        self.order_owner = order_owner
        self.pay_gem = pay_gem
        self.pay_amt = pay_amt
        self.paid_amt = paid_amt
        self.buy_gem = buy_gem
        self.buy_amt = buy_amt
        self.bought_amt = bought_amt
        self.open = open

    @staticmethod
    def get_fields(offer_query: Any) -> List:
        """Helper method to build a list of fields for the offers subgraph entity."""
        return [
            offer_query.id,
            offer_query.timestamp,
            offer_query.index,
            offer_query.maker.id,
            offer_query.from_address.id,
            offer_query.pay_gem,
            offer_query.buy_gem,
            offer_query.pay_amt,
            offer_query.buy_amt,
            offer_query.paid_amt,
            offer_query.bought_amt,
            offer_query.price,
            offer_query.open,
            offer_query.removed_timestamp,
            offer_query.removed_block,
            offer_query.transaction.id,
            offer_query.transaction.block_number,
            offer_query.transaction.block_index,
            offer_query.pay_amt_decimals,
            offer_query.buy_amt_decimals,
            offer_query.paid_amt_decimals,
            offer_query.bought_amt_decimals,
            offer_query.pay_gem_symbol,
            offer_query.buy_gem_symbol,
            offer_query.datetime,
        ]


class SubgraphTrade:
    """Helper object for querying subgraph Trades"""

    @staticmethod
    def get_fields(trade_query: Any) -> List:
        """Helper method to build a list of fields for the offers subgraph entity."""
        return [
            trade_query.id,
            trade_query.timestamp,
            trade_query.take_gem,
            trade_query.give_gem,
            trade_query.take_amt,
            trade_query.give_amt,
            trade_query.taker.id,
            trade_query.from_address.id,
            trade_query.transaction.block_number,
            trade_query.transaction.block_index,
            trade_query.index,
            trade_query.offer.maker.id,
            trade_query.offer.from_address.id,
            trade_query.offer.transaction.block_number,
            trade_query.offer.transaction.block_index,
            trade_query.offer.index,
            trade_query.take_amt_decimals,
            trade_query.give_amt_decimals,
            trade_query.take_gem_symbol,
            trade_query.datetime,
        ]



import os
import logging
from concurrent.futures import ThreadPoolExecutor
from enum import Enum
from typing import Optional, Dict, Union

import yaml
from eth_typing import ChecksumAddress
from web3 import Web3

from rubi.contracts import ERC20, RubiconMarket, RubiconRouter, TransactionHandler

# from rubi.data import MarketData

logger = logging.getLogger(__name__)


class NetworkId(Enum):
    # MAINNET
    OPTIMISM = 10
    ARBITRUM_ONE = 42161
    BASE = 8453

    # TESTNET
    OPTIMISM_GOERLI = 420
    ARBITRUM_GOERLI = 421613
    POLYGON_MUMBAI = 80001


class Network:
    """This class represents a network and can be used to instantiate contracts using the `from_network` method. It
    provides and easy way to configure a client or contracts.
    """

    def __init__(
        self,
        w3: Web3,
        # The below all come from network_config/{network_name}/network.yaml
        name: str,
        chain_id: int,
        currency: str,
        rpc_url: str,
        explorer_url: str,
        market_data_url: str,
        market_data_fallback_url: str,
        rubicon: Dict,
        token_addresses: Dict,
        # optional custom token config file from the user
        custom_token_addresses_file: Optional[str] = None,
    ):
        """Initializes a Network instance.

        :param w3: The Web3 instance connected to the network.
        :type w3: Web3
        :param name: The name of the network.
        :type name: str
        :param chain_id: The chain ID of the network.
        :type chain_id: int
        :param currency: The base currency of the network.
        :type currency: str
        :param rpc_url: The RPC URL of the network.
        :type rpc_url: str
        :param explorer_url: The URL of the network explorer.
        :type explorer_url: str
        :param market_data_url: the URL of the market data subgraph (RubiconV2)
        :param rubicon: Dictionary containing Rubicon contract parameters.
        :type rubicon: dict
        :param token_addresses: Dictionary containing token addresses on the network.
        :type token_addresses: dict
        :param custom_token_addresses_file: The name of a yaml file (relative to the current working directory) with
            custom token addresses. Overwrites the token config found in network_config/{chain}/network.yaml.
            (optional, default is None).
        :type custom_token_addresses_file: Optional[str]
        """
        # General config
        self.name = name
        self.w3 = w3
        self.chain_id = chain_id
        self.currency = currency
        self.rpc_url = rpc_url
        self.explorer_url = explorer_url

        # Rubicon contracts
        self.rubicon_market = RubiconMarket.from_address(
            w3=self.w3, address=rubicon["market"]
        )
        self.rubicon_router = RubiconRouter.from_address(
            w3=self.w3, address=rubicon["router"]
        )

        # Tokens
        custom_token_addresses = self._custom_token_addresses(
            custom_token_addresses_file=custom_token_addresses_file
        )
        token_addresses.update(custom_token_addresses)

        checksummed_token_addresses = self._checksum_token_addresses(
            token_addresses=token_addresses
        )

        futures = {}
        with ThreadPoolExecutor() as executor:
            for name, address in checksummed_token_addresses.items():
                future = executor.submit(ERC20.from_address, self.w3, address)
                futures[name] = future

        self.tokens: Dict[Union[ChecksumAddress, str], ERC20] = {}
        for name, future in futures.items():
            erc20 = future.result()

            self.tokens[name] = erc20
            self.tokens[erc20.address] = erc20

        # Transaction Handler
        self.transaction_handler = TransactionHandler(
            w3=self.w3,
            contracts=[
                self.rubicon_market.contract,
                self.rubicon_router.contract,
                # We only need one ERC20 contract in order to decode ERC20 events
                list(self.tokens.values())[0].contract,
            ],
        )

        # Subgraph urls
        # TODO: currently we are utilizing just a single url, we should switch to a dictionary as the number of
        #  subgraphs we support grows
        self.market_data_url = market_data_url
        self.market_data_fallback_url = market_data_fallback_url

    @classmethod
    def from_http_node_url(
        cls,
        http_node_url: str,
        custom_token_addresses_file: Optional[str] = None,
    ) -> "Network":
        """Create a Network instance based on the node url provided. A call is then made to this node to get the
        chain_id which links to network_config/{network_name}/ using the NetworkId Enum.

        :param http_node_url: The URL of the HTTP node for the network.
        :type http_node_url: str
        :param custom_token_addresses_file: The name of a yaml file (relative to the current working directory) with
            custom token addresses. Overwrites the token config found in network_config/{chain}/network.yaml.
            (optional, default is None).
        :type custom_token_addresses_file: Optional[str]
        :return: A Network instance based on the network configuration.
        :rtype: Network
        :raises Exception: If no network configuration file is found for the specified network name.
        """
        w3 = Web3(Web3.HTTPProvider(http_node_url))

        network_name = NetworkId(w3.eth.chain_id).name.lower()

        try:
            path = f"{os.path.dirname(os.path.abspath(__file__))}/../../network_config/{network_name}"

            with open(f"{path}/network.yaml") as f:
                network_data = yaml.safe_load(f)
                return cls(
                    w3=w3,
                    custom_token_addresses_file=custom_token_addresses_file,
                    **network_data,
                )
        except FileNotFoundError:
            raise Exception(
                f"No network config found for {network_name}. There should be a corresponding folder in "
                f"the network_config directory."
            )

    @staticmethod
    def _custom_token_addresses(custom_token_addresses_file: str) -> Dict[str, str]:
        if not custom_token_addresses_file:
            return {}
        if not (
            custom_token_addresses_file.endswith(".yaml")
            or custom_token_addresses_file.endswith(".yml")
        ):
            raise Exception(
                f"token_config_file: {custom_token_addresses_file} must be a yaml file."
            )

        working_directory = os.getcwd()
        try:
            with open(f"{working_directory}/{custom_token_addresses_file}") as f:
                custom_token_addresses = yaml.safe_load(f)

                return custom_token_addresses
        except FileNotFoundError:
            raise Exception(
                f"could not find token_config_file, expected it to be a yaml file here: "
                f"{working_directory}/{custom_token_addresses_file}."
            )
        except AttributeError:
            raise Exception(
                f"{custom_token_addresses_file} cannot be empty, should be in the format: "
                f"token_symbol: address, e.g. WETH: chain_specific_weth_address"
            )

    def _checksum_token_addresses(
        self, token_addresses: Dict[str, str]
    ) -> Dict[str, ChecksumAddress]:
        checksummed_token_addresses: dict[str, ChecksumAddress] = {}
        for k, v in token_addresses.items():
            checksummed_token_addresses[k] = self.w3.to_checksum_address(v)

        return checksummed_token_addresses

    def token_from_address(self, address: Union[ChecksumAddress, str]):
        try:
            address = self.w3.to_checksum_address(address)
        except:
            logger.error(f"Could not checksum address {address}")
            return

        try:
            erc20 = ERC20.from_address(self.w3, address)
        except:
            logger.error(f"Could not find token with address {address}")
            return

        self.tokens[erc20.symbol] = erc20
        self.tokens[erc20.address] = erc20



import logging
from _decimal import Decimal
from datetime import datetime
from typing import Optional, Dict, List, Union

import pandas as pd
from eth_typing import ChecksumAddress
from subgrounds import Subgrounds, Subgraph, SyntheticField
from subgrounds.pagination import ShallowStrategy
from web3 import Web3

from rubi.contracts import ERC20
from rubi.data.helpers import QueryValidation
from rubi.data.helpers import SubgraphOffer, SubgraphTrade
from rubi.network import Network

logger = logging.getLogger(__name__)

# Stop subgrounds from logging kak
logging.getLogger("subgrounds").setLevel(logging.WARNING)


class MarketData:
    """This class represents the RubiconV2 Subgraph, which contains data primarily related to the RubiconMarket.sol
    contract.

    :param url: a RubiconV2 Subgraph url endpoint that should be utilized for this class
    :type url: str
    :param fallback_url: a fallback url if we fail to connect to the primary url
    :type fallback_url: str
    :param tokens: A dictionary of token addresses and symbols to their ERC20 instantiation
    :type tokens: Dict[ChecksumAddress | str, ERC20]
    """

    def __init__(
        self,
        url: str,
        fallback_url: str,
        network: Optional[Network] = None,
    ):
        """constructor method"""

        self.subgrounds = Subgrounds()

        if network is not None:
            self.tokens = network.tokens
            self.netwrok = network
        else:
            self.tokens = None  # noqa
            self.network = None  # noqa

        self.subgraph: Subgraph = self._initialize_subgraph(
            url=url, fallback_url=fallback_url
        )

        # Subgraph objects
        self.offer = self._initialize_subgraph_offer()
        self.trade = self._initialize_subgraph_trade()

    @classmethod
    def from_network(
        cls,
        network: Network,
    ):
        """Initialize the MarketData object from a network object

        :param network: The network object
        :type network: Network
        """

        return cls(
            url=network.market_data_url,
            fallback_url=network.market_data_fallback_url,
            network=network,
        )

    def _initialize_subgraph(
        self, url: str, fallback_url: str, attempts: int = 3
    ) -> Subgraph:
        """Initialize the subgraph

        :param url: The subgraph url
        :type url: str
        :param fallback_url: The fallback subgraph url
        :type url: str
        :param attempts: The number of connection attempts to make
        :type attempts: int
        :return: A initialized subgraph instance
        :rtype: Subgraph
        """

        subgraph = None

        for attempt in range(attempts):
            try:
                # TODO: we should add a check here to guarantee the schema matches what we expect to be receiving
                subgraph = self.subgrounds.load_subgraph(url=url)
                break
            except Exception as e:
                logger.debug(f"Exception loading subgraph: {e}")
                continue

        if subgraph is None:
            for attempt in range(attempts):
                try:
                    subgraph = self.subgrounds.load_subgraph(url=fallback_url)
                    break
                except Exception as e:
                    logger.debug(f"Exception loading subgraph: {e}")
                    continue

        if subgraph is None:
            raise ValueError(
                f"Both subgraph_url: {url} and fallback_url: {fallback_url} failed when attempting to load."
            )

        return subgraph

    ######################################################################
    # Subgraph objects
    ######################################################################

    def _erc20_to_decimal(
        self, gem: Union[ChecksumAddress, str], amt: int
    ) -> Optional[Decimal]:
        """Helper to convert an amount to decimals for the given ERC20"""

        if self.tokens.get(Web3.to_checksum_address(gem)) is None:
            self.network.token_from_address(gem)

        try:
            return self.tokens[Web3.to_checksum_address(gem)].to_decimal(amt)
        except KeyError:
            return None

    def _erc20_to_symbol(self, gem: Union[ChecksumAddress, str]) -> Optional[str]:
        """Helper to get the symbol of the given ERC20"""

        if self.tokens.get(Web3.to_checksum_address(gem)) is None:
            self.network.token_from_address(gem)

        try:
            return self.tokens[Web3.to_checksum_address(gem)].symbol
        except KeyError:
            return None

    def _initialize_subgraph_offer(self):
        """Initialize the Subgraph offer object and add synthetic fields"""

        offer = self.subgraph.Offer  # noqa

        offer.datetime = SyntheticField(
            f=lambda timestamp: str(datetime.fromtimestamp(timestamp)),
            type_=SyntheticField.STRING,
            deps=[offer.timestamp],
        )

        if self.tokens is not None:
            offer.pay_amt_decimals = SyntheticField(
                f=lambda pay_amt, pay_gem: self._erc20_to_decimal(
                    gem=pay_gem, amt=pay_amt
                ),
                type_=SyntheticField.FLOAT,
                deps=[offer.pay_amt, offer.pay_gem],
            )

            offer.buy_amt_decimals = SyntheticField(
                f=lambda buy_amt, buy_gem: self._erc20_to_decimal(
                    gem=buy_gem, amt=buy_amt
                ),
                type_=SyntheticField.FLOAT,
                deps=[offer.buy_amt, offer.buy_gem],
            )

            offer.paid_amt_decimals = SyntheticField(
                f=lambda paid_amt, pay_gem: self._erc20_to_decimal(
                    gem=pay_gem, amt=paid_amt
                ),
                type_=SyntheticField.FLOAT,
                deps=[offer.paid_amt, offer.pay_gem],
            )

            offer.bought_amt_decimals = SyntheticField(
                f=lambda bought_amt, buy_gem: self._erc20_to_decimal(
                    gem=buy_gem, amt=bought_amt
                ),
                type_=SyntheticField.FLOAT,
                deps=[offer.bought_amt, offer.buy_gem],
            )

            offer.pay_gem_symbol = SyntheticField(
                f=lambda pay_gem: self._erc20_to_symbol(gem=pay_gem),
                type_=SyntheticField.STRING,
                deps=[offer.pay_gem],
            )

            offer.buy_gem_symbol = SyntheticField(
                f=lambda buy_gem: self._erc20_to_symbol(gem=buy_gem),
                type_=SyntheticField.STRING,
                deps=[offer.buy_gem],
            )

        return offer

    def _initialize_subgraph_trade(self):
        """Initialize the Subgraph trade object and add synthetic fields"""

        take = self.subgraph.Take  # noqa

        take.datetime = SyntheticField(
            f=lambda timestamp: str(datetime.fromtimestamp(timestamp)),
            type_=SyntheticField.STRING,
            deps=[take.timestamp],
        )

        if self.tokens is not None:
            take.take_amt_decimals = SyntheticField(
                f=lambda take_amt, take_gem: self._erc20_to_decimal(
                    gem=take_gem, amt=take_amt
                ),
                type_=SyntheticField.FLOAT,
                deps=[take.take_amt, take.take_gem],
            )

            take.give_amt_decimals = SyntheticField(
                f=lambda give_amt, give_gem: self._erc20_to_decimal(
                    gem=give_gem, amt=give_amt
                ),
                type_=SyntheticField.FLOAT,
                deps=[take.give_amt, take.give_gem],
            )

            take.take_gem_symbol = SyntheticField(
                f=lambda take_gem: self._erc20_to_symbol(gem=take_gem),
                type_=SyntheticField.STRING,
                deps=[take.take_gem],
            )

            take.give_gem_symbol = SyntheticField(
                f=lambda give_gem: self._erc20_to_symbol(gem=give_gem),
                type_=SyntheticField.STRING,
                deps=[take.give_gem],
            )

        return take

    ######################################################################
    # Subgraph Query Methods
    ######################################################################

    # TODO: refactor using a decorator to handle the parameter validation
    def get_offers(
        self,
        maker: Optional[ChecksumAddress] = None,
        from_address: Optional[ChecksumAddress] = None,
        pay_gem: Optional[ChecksumAddress] = None,
        buy_gem: Optional[ChecksumAddress] = None,
        side: Optional[str] = None,
        open: Optional[bool] = None,
        start_time: Optional[int] = None,
        end_time: Optional[int] = None,
        start_block: Optional[int] = None,  # TODO: add in start_block and end_block
        end_block: Optional[int] = None,  # TODO: add in start_block and end_block
        first: int = 1000,
        # TODO: expand order_by options
        order_by: str = "timestamp",
        order_direction: str = "desc",
        as_dataframe: bool = True,
    ) -> Optional[pd.DataFrame] | List[SubgraphOffer]:
        """Returns a dataframe of offers placed on the market contract, with the option to pass in filters.

        :param maker: the address of the maker of the offer
        :type maker: Optional[ChecksumAddress]
        :param from_address: the address that originated the transaction that created the offer
        :type from_address: Optional[ChecksumAddress]
        :param pay_gem: the address of the token that the maker is offering (will override pair_name if both are passed)
        :type pay_gem: Optional[ChecksumAddress]
        :param buy_gem: the address of the token that the maker is requesting (will override pair_name if both are passed)
        :type buy_gem: Optional[ChecksumAddress]
        :param side: The side we are querying for
        :type side: Optional[str]
        :param open: whether the offer is still active
        :type open:Optional[bool]
        :param start_time: the timestamp of the earliest offer to return
        :type start_time: int
        :param end_time: the timestamp of the latest offer to return
        :type end_time: int
        :param first: the number of offers to return
        :type first: int
        :param order_by: the field to order the offers by (default: timestamp, options: timestamp, price)
        :type order_by: str
        :param order_direction: the direction to order the offers by (default: desc, options: asc, desc)
        :type order_direction: str
        :param as_dataframe: If the response should be a dataframe (default: True)
        :type as_dataframe: bool
        :return: a dataframe of offers placed on the market contract or a list of subgraph offer objects
        :rtype: Optional[pd.DataFrame] | List[SubgraphOffer]
        """

        offer_query = self._build_offers_query(
            order_by=order_by,
            order_direction=order_direction,
            first=first,
            maker=maker,
            from_address=from_address,
            pay_gem=pay_gem,
            buy_gem=buy_gem,
            open=open,
            start_time=start_time,
            end_time=end_time,
            start_block=start_block,
            end_block=end_block,
        )

        query_fields = SubgraphOffer.get_fields(offer_query=offer_query)
        if as_dataframe:
            response = self._query_offers_as_dataframe(query_fields=query_fields)
            # TODO: we could also pass this data to the offers_query method and handle it there, could help with price
            if response is not None and not response.empty:
                response["side"] = side if side else "N/A"

            return response
        else:
            response = self._query_offers(query_fields=query_fields)

            return response

    def get_trades(
        self,
        # TODO: resolve #63 and add in conditional filter for offer maker/from_address
        # maker: Optional[Union[ChecksumAddress, str]],
        taker: Optional[Union[ChecksumAddress, str]] = None,
        from_address: Optional[Union[ChecksumAddress, str]] = None,
        take_gem: Optional[Union[ChecksumAddress, str]] = None,
        give_gem: Optional[Union[ChecksumAddress, str]] = None,
        side: Optional[str] = None,
        start_time: Optional[int] = None,
        end_time: Optional[int] = None,
        start_block: Optional[int] = None,
        end_block: Optional[int] = None,
        maker: Optional[
            Union[ChecksumAddress, str]
        ] = None,  # TODO: implement this with nested filtering
        maker_from_address: Optional[
            Union[ChecksumAddress, str]
        ] = None,  # TODO: implement this with nested filtering
        first: int = 1000,
        # TODO: expand order_by options
        order_by: str = "timestamp",
        order_direction: str = "desc",
    ) -> pd.DataFrame:
        """Returns a dataframe of trades that have occurred on the market contract, with the option to pass in filters.

        :param taker: the address of the taker of the trade
        :type taker: str
        :param from_address: the address that originated the transaction that created the trade
        :type from_address: str
        :param take_gem: the address of the token that the taker received
        :type take_gem: str
        :param give_gem: the address of the token that the taker gave
        :type give_gem: str
        :param side: The side we are querying for
        :type side: Optional[str]
        :param start_time: the timestamp of the earliest trade to return
        :type start_time: int
        :param end_time: the timestamp of the latest trade to return
        :type end_time: int
        :param first: the number of trades to return
        :type first: int
        :param order_by: the field to order the trades by (default: timestamp, options: timestamp)
        :type order_by: str
        :param order_direction: the direction to order the trades by (default: desc, options: asc, desc)
        :type order_direction: str
        :return: a dataframe of trades that have occurred on the market contract
        :rtype: pd.DataFrame
        """

        trade_query = self._build_trades_query(
            order_by=order_by,
            order_direction=order_direction,
            first=first,
            taker=taker,
            from_address=from_address,
            take_gem=take_gem,
            give_gem=give_gem,
            start_time=start_time,
            end_time=end_time,
            start_block=start_block,
            end_block=end_block,
            maker=maker,
            maker_from_address=maker_from_address,
        )
        query_fields = SubgraphTrade.get_fields(trade_query=trade_query)
        df = self._query_trades_as_dataframe(query_fields=query_fields)
        if df is not None and not df.empty:
            df["side"] = side if side else "N/A"

        return df

    ######################################################################
    # helper methods
    ######################################################################

    def _build_offers_query(
        self,
        order_by: str,
        order_direction: str,
        first: int,
        maker: Optional[Union[ChecksumAddress, str]] = None,
        from_address: Optional[Union[ChecksumAddress, str]] = None,
        pay_gem: Optional[Union[ChecksumAddress, str]] = None,
        buy_gem: Optional[Union[ChecksumAddress, str]] = None,
        open: Optional[bool] = None,
        start_time: Optional[int] = None,
        end_time: Optional[int] = None,
        start_block: Optional[int] = None,  # TODO: add in start_block and end_block
        end_block: Optional[int] = None,  # TODO: add in start_block and end_block
    ):
        """Helper method build an offers query."""

        QueryValidation.validate_offer_query(
            order_by=order_by,
            order_direction=order_direction,
            first=first,
        )

        if order_by.lower() == "timestamp":
            order_by = self.offer.timestamp
        elif order_by.lower() == "price":
            order_by = self.offer.price

        order_direction = order_direction.lower()

        # build the list of where conditions
        where = [
            self.offer.maker == str(maker).lower() if maker else None,
            self.offer.from_address == str(from_address).lower()
            if from_address
            else None,
            self.offer.pay_gem == str(pay_gem).lower() if pay_gem else None,
            self.offer.buy_gem == str(buy_gem).lower() if buy_gem else None,
            self.offer.open == open if open is not None else None,
            self.offer.timestamp >= start_time if start_time else None,
            self.offer.timestamp <= end_time if end_time else None,
        ]
        where = [condition for condition in where if condition is not None]

        offers_query = self.subgraph.Query.offers(  # noqa
            orderBy=order_by,
            orderDirection=order_direction,
            first=first,
            where=where if where else {},
        )

        return offers_query

    def _build_trades_query(
        self,
        order_by: str,
        order_direction: str,
        first: int,
        taker: Optional[Union[ChecksumAddress, str]] = None,
        from_address: Optional[Union[ChecksumAddress, str]] = None,
        take_gem: Optional[Union[ChecksumAddress, str]] = None,
        give_gem: Optional[Union[ChecksumAddress, str]] = None,
        start_time: Optional[int] = None,
        end_time: Optional[int] = None,
        start_block: Optional[int] = None,
        end_block: Optional[int] = None,
        maker: Optional[
            Union[ChecksumAddress, str]
        ] = None,  # TODO: implement this with nested filtering
        maker_from_address: Optional[
            Union[ChecksumAddress, str]
        ] = None,  # TODO: implement this with nested filtering
    ):
        """Helper method build a trades query."""

        QueryValidation.validate_trade_query(
            order_by=order_by,
            order_direction=order_direction,
            first=first,
        )

        if order_by.lower() == "timestamp":
            order_by = self.trade.timestamp

        order_direction = order_direction.lower()

        # build the list of where conditions
        where = [
            self.trade.taker == str(taker).lower() if taker else None,
            self.trade.from_address == str(from_address).lower()
            if from_address
            else None,
            self.trade.take_gem == str(take_gem).lower() if take_gem else None,
            self.trade.give_gem == str(give_gem).lower() if give_gem else None,
            self.trade.timestamp >= start_time if start_time else None,
            self.trade.timestamp <= end_time if end_time else None,
        ]
        where = [condition for condition in where if condition is not None]

        trades_query = self.subgraph.Query.takes(  # noqa
            orderBy=order_by,
            orderDirection=order_direction,
            first=first,
            where=where if where else {},
        )

        return trades_query

    def _query_offers_as_dataframe(self, query_fields: List) -> Optional[pd.DataFrame]:
        """Helper method to query the offers subgraph entity and return a dataframe."""

        df = self.subgrounds.query_df(
            query_fields,
            # TODO: maybe we give the user the option to define a custom pagination strategy.
            pagination_strategy=ShallowStrategy,  # noqa
        )

        if df.empty:
            return df

        df.columns = [col.replace("offers_", "") for col in df.columns]
        df.columns = [col.replace("_id", "") for col in df.columns]

        # convert the id to an integer
        df["id"] = df["id"].apply(lambda x: int(x, 16))

        # TODO: apply any data type conversions to the dataframe - possibly converting unformatted values to integers
        return df

    def _query_offers(self, query_fields: List) -> List[SubgraphOffer]:
        """Helper method to query the offers subgraph entity."""

        response = self.subgrounds.query_json(
            query_fields,
            # TODO: maybe we give the user the option to define a custom pagination strategy.
            pagination_strategy=ShallowStrategy,  # noqa
        )

        if response:
            raw_offers = list(response[0].values())[0]
            offers: List[SubgraphOffer] = []

            for raw_offer in raw_offers:
                offers.append(
                    SubgraphOffer(
                        order_id=int(raw_offer["id"], 16),
                        order_owner=Web3.to_checksum_address(raw_offer["maker"]["id"]),
                        pay_gem=Web3.to_checksum_address(raw_offer["pay_gem"]),
                        pay_amt=raw_offer["pay_amt"],
                        paid_amt=raw_offer["paid_amt"],
                        buy_gem=Web3.to_checksum_address(raw_offer["buy_gem"]),
                        buy_amt=raw_offer["buy_amt"],
                        bought_amt=raw_offer["bought_amt"],
                        open=raw_offer["open"],
                    )
                )
            return offers

        return []

    def _query_trades_as_dataframe(
        self,
        query_fields: List,
    ) -> Optional[pd.DataFrame]:
        """Helper method to query the trades subgraph entity."""

        df = self.subgrounds.query_df(
            query_fields,
            # TODO: maybe we give the user the option to define a custom pagination strategy.
            pagination_strategy=ShallowStrategy,  # noqa
        )

        if df.empty:
            return df

        df.columns = [col.replace("takes_", "") for col in df.columns]
        df.columns = [col.replace("_id", "") for col in df.columns]

        return df



from _decimal import Decimal
from typing import Optional

from eth_typing import ChecksumAddress


class Approval:
    """Class representing an ERC20 approval.

    :param token: The token being approved
    :type token: str
    :param amount: The approval amount
    :type amount: Decimal
    :param spender: The spender being approved
    :type: spender: Optional[ChecksumAddress]
    """

    def __init__(self, token: str, amount: Decimal, spender: Optional[ChecksumAddress]):
        """constructor method."""
        self.token = token
        self.amount = amount
        self.spender = spender

    def __repr__(self):
        items = ("{}={!r}".format(k, self.__dict__[k]) for k in self.__dict__)
        return "{}({})".format(type(self).__name__, ", ".join(items))


class RubiconMarketApproval(Approval):
    """A subclass for a RubiconMarket approval"""

    def __init__(self, token: str, amount: Decimal):
        """constructor method."""
        super().__init__(token=token, amount=amount, spender=None)


class RubiconRouterApproval(Approval):
    """A subclass for a RubiconRouter approval"""

    def __init__(self, token: str, amount: Decimal):
        """constructor method."""
        super().__init__(token=token, amount=amount, spender=None)


class ApprovalEvent(Approval):
    """A Subclass for approval events."""

    def __init__(
        self,
        token: str,
        amount: Decimal,
        spender: ChecksumAddress,
        source: ChecksumAddress,
    ):
        """constructor method."""
        super().__init__(token=token, amount=amount, spender=spender)

        self.source = source


class Transfer:
    """Class representing an ERC20 transfer.

    :param token: The token being transferred
    :type token: str
    :param amount: The transfer amount
    :type amount: Decimal
    :param recipient: The recipient of the transfer
    :type: recipient: ChecksumAddress
    """

    def __init__(self, token: str, amount: Decimal, recipient: ChecksumAddress):
        """constructor method."""
        self.token = token
        self.amount = amount
        self.recipient = recipient

    def __repr__(self):
        items = ("{}={!r}".format(k, self.__dict__[k]) for k in self.__dict__)
        return "{}({})".format(type(self).__name__, ", ".join(items))


class TransferEvent(Transfer):
    """A Subclass for transfer events."""

    def __init__(
        self,
        token: str,
        amount: Decimal,
        recipient: ChecksumAddress,
        source: ChecksumAddress,
    ):
        """constructor method."""
        super().__init__(token=token, amount=amount, recipient=recipient)

        self.source = source



from _decimal import Decimal
from enum import Enum
from typing import Optional, Union

from eth_typing import ChecksumAddress
from web3 import Web3

from rubi.data.helpers import SubgraphOffer
from rubi.contracts import (
    ERC20,
    BaseEvent,
    EmitOfferEvent,
    EmitCancelEvent,
    EmitTakeEvent,
    EmitDeleteEvent,
    EmitFeeEvent,
)


class OrderSide(Enum):
    """Enumeration representing the order side."""

    BUY = "BUY"
    SELL = "SELL"
    NEUTRAL = "NEUTRAL"

    def sign(self) -> int:
        """
        :return: Numerical value of the side.
        :rtype: int
        """
        match self:
            case OrderSide.NEUTRAL:
                return 0

            case OrderSide.BUY:
                return 1

            case OrderSide.SELL:
                return -1

    def opposite(self) -> "OrderSide":
        match self:
            case OrderSide.NEUTRAL:
                return OrderSide.NEUTRAL

            case OrderSide.BUY:
                return OrderSide.SELL

            case OrderSide.SELL:
                return OrderSide.BUY


class OrderType(Enum):
    """Enumeration representing the order type."""

    MARKET = "MARKET"
    LIMIT = "LIMIT"

    # Only used for events coming from the RubiconMarket
    LIMIT_TAKEN = "LIMIT_TAKEN"
    LIMIT_DELETED = "LIMIT_DELETED"
    CANCEL = "CANCEL"


class BaseOrder:
    """Base class for representing a new order.

    :param pair_name: The name of the trading pair.
    :type pair_name: str
    :param order_type: The type of the order.
    :type order_type: OrderType
    :param order_side: The side of the order (buy or sell).
    :type order_side: OrderSide
    """

    def __init__(
        self,
        pair_name: str,
        order_type: OrderType,
        order_side: OrderSide,
    ):
        """constructor method."""
        self.pair_name = pair_name
        self.order_side = order_side
        self.order_type = order_type


class NewMarketOrder(BaseOrder):
    """Class representing a new market order.

    :param pair_name: The name of the pair being traded e.g. WETH/USDC.
    :type pair_name: str
    :param order_side: The side of the order (BUY or SELL).
    :type order_side: OrderSide
    :param size: The size of the order.
    :type size: Decimal
    :param worst_execution_price: The worst execution price for the order (optional). Defaults to 0 if selling and
        10 million if buying as random bounds.
    :type worst_execution_price: Decimal
    """

    def __init__(
        self,
        pair_name: str,
        order_side: OrderSide,
        size: Decimal,
        worst_execution_price: Optional[Decimal],
    ):
        """constructor method."""

        super().__init__(
            pair_name=pair_name,
            order_type=OrderType.MARKET,
            order_side=order_side,
        )
        self.size = size

        if worst_execution_price is None:
            self.worst_execution_price = (
                Decimal("0") if order_side.SELL else Decimal("10") ** Decimal("7")
            )
        else:
            self.worst_execution_price = worst_execution_price


class NewLimitOrder(BaseOrder):
    """Class representing a new limit order

    :param pair_name: The name of the pair being traded e.g. WETH/USDC.
    :type pair_name: str
    :param order_side: The side of the order (buy or sell).
    :type order_side: OrderSide
    :param size: The size of the order.
    :type size: Decimal
    :param price: The price of the order.
    :type price: Decimal
    """

    def __init__(
        self, pair_name: str, order_side: OrderSide, size: Decimal, price: Decimal
    ):
        """constructor method."""
        super().__init__(
            pair_name=pair_name,
            order_type=OrderType.LIMIT,
            order_side=order_side,
        )

        self.size = size
        self.price = price

    def __repr__(self):
        items = ("{}={!r}".format(k, self.__dict__[k]) for k in self.__dict__)
        return "{}({})".format(type(self).__name__, ", ".join(items))


class UpdateLimitOrder(BaseOrder):
    """Class representing an update to an existing limit order

    :param pair_name: The name of the pair being traded e.g. WETH/USDC.
    :type pair_name: str
    :param order_side: The side of the order (BUY or SELL).
    :type order_side: OrderSide
    :param order_id: The ID of the order to update.
    :type order_id: int
    :param size: The updated size of the order.
    :type size: Decimal
    :param price: The updated price of the order.
    :type price: Decimal
    """

    def __init__(
        self,
        pair_name: str,
        order_side: OrderSide,
        order_id: int,
        size: Decimal,
        price: Decimal,
    ):
        """constructor method"""
        super().__init__(
            pair_name=pair_name,
            order_type=OrderType.LIMIT,
            order_side=order_side,
        )

        self.order_id = order_id
        self.size = size
        self.price = price


class LimitOrder(BaseOrder):
    """Class representing an existing limit order

    :param pair_name: The name of the pair being traded e.g. WETH/USDC.
    :type pair_name: str
    :param order_side: The side of the order (BUY or SELL).
    :type order_side: OrderSide
    :param order_id: The ID of the order to update.
    :type order_id: int
    :param order_owner: The address of the owner of the order
    :type order_owner: ChecksumAddress
    :param size: The updated size of the order.
    :type size: Decimal
    :param price: The updated price of the order.
    :type price: Decimal
    :param filled_size: The amount of the order that has been filled
    :type filled_size: Decimal
    """

    def __init__(
        self,
        pair_name: str,
        order_side: OrderSide,
        order_id: int,
        order_owner: ChecksumAddress,
        size: Decimal,
        price: Decimal,
        filled_size: Decimal = Decimal("0"),
        open: bool = True,
    ):
        """constructor method"""
        super().__init__(
            pair_name=pair_name,
            order_type=OrderType.LIMIT,
            order_side=order_side,
        )

        self.order_id = order_id
        self.order_owner = order_owner
        self.size = size
        self.price = price
        self.filled_size = filled_size
        self.open = open

    @property
    def remaining_size(self) -> Decimal:
        return self.size - self.filled_size

    @classmethod
    def from_order_event(cls, order_event: "OrderEvent") -> "LimitOrder":
        if order_event.order_type != OrderType.LIMIT:
            raise Exception("event must have LIMIT as order_type.")

        return cls(
            pair_name=order_event.pair_name,
            order_side=order_event.order_side,
            order_id=order_event.limit_order_id,
            order_owner=order_event.limit_order_owner,
            size=order_event.size,
            price=order_event.price,
        )

    @classmethod
    def from_subgraph_offer(
        cls, base_asset: ERC20, quote_asset: ERC20, offer: SubgraphOffer
    ) -> "LimitOrder":
        if base_asset.address == offer.buy_gem:
            size = base_asset.to_decimal(offer.buy_amt)
            price = quote_asset.to_decimal(offer.pay_amt) / size
            filled_size = base_asset.to_decimal(offer.bought_amt)

            return cls(
                pair_name=f"{base_asset.symbol}/{quote_asset.symbol}",
                order_side=OrderSide.BUY,
                order_id=offer.order_id,
                order_owner=offer.order_owner,
                size=size,
                price=price,
                filled_size=filled_size,
            )
        else:
            size = base_asset.to_decimal(offer.pay_amt)
            price = quote_asset.to_decimal(offer.buy_amt) / size
            filled_size = base_asset.to_decimal(offer.paid_amt)

            return cls(
                pair_name=f"{base_asset.symbol}/{quote_asset.symbol}",
                order_side=OrderSide.SELL,
                order_id=offer.order_id,
                order_owner=offer.order_owner,
                size=size,
                price=price,
                filled_size=filled_size,
            )

    def update_with_take(self, order_event: "OrderEvent"):
        self.filled_size += order_event.size

    def __repr__(self):
        items = ("{}={!r}".format(k, self.__dict__[k]) for k in self.__dict__)
        return "{}({})".format(type(self).__name__, ", ".join(items))


class NewCancelOrder(BaseOrder):
    """Class representing a limit order cancellation

    :param pair_name: The name of the trading pair.
    :type pair_name: str
    :param order_id: The ID of the order to cancel.
    :type order_id: int
    """

    def __init__(self, pair_name: str, order_id: int):
        """constructor method"""
        super().__init__(
            pair_name=pair_name,
            order_type=OrderType.LIMIT,
            order_side=OrderSide.NEUTRAL,
        )

        self.order_id = order_id


class OrderEvent:
    """Class to represent Rubicon Market events as an order

    :param limit_order_id: The ID of the limit order.
    :type limit_order_id: int
    :param limit_order_owner: The owner of the limit order.
    :type limit_order_owner: ChecksumAddress
    :param market_order_owner: The owner of the market order (optional). Only has a value if event is an emitTakeEvent.
    :type market_order_owner: Optional[ChecksumAddress]
    :param pair_name: The name of the pair being traded e.g. WETH/USDC.
    :type pair_name: str
    :param order_side: The side of the order (BUY or SELL).
    :type order_side: OrderSide
    :param order_type: The type of the order (MARKET, LIMIT, LIMIT_TAKEN, LIMIT_DELETED, or CANCEL).
    :type order_type: OrderType
    :param price: The price of the order.
    :type price: Decimal
    :param size: The size of the order.
    :type size: Decimal
    """

    def __init__(
        self,
        limit_order_id: int,
        limit_order_owner: ChecksumAddress,
        market_order_owner: Optional[ChecksumAddress],
        pair_name: str,
        order_side: Optional[OrderSide],
        order_type: OrderType,
        price: Optional[Decimal],
        size: Optional[Decimal],
    ):
        self.limit_order_id = limit_order_id
        self.limit_order_owner = limit_order_owner
        self.market_order_owner: Optional[ChecksumAddress] = market_order_owner
        self.pair_name = pair_name
        self.order_side = order_side
        self.order_type = order_type
        self.price = price
        self.size = size

    @classmethod
    def from_event(
        cls,
        base_asset: ERC20,
        quote_asset: ERC20,
        event: BaseEvent,
        wallet: ChecksumAddress,
    ) -> "OrderEvent":
        """Create an OrderEvent from a BaseEvent emitted by the Rubicon Market.

        :param base_asset: The base asset associated with the event.
        :type base_asset: ERC20
        :param quote_asset: The quote asset associated with the event.
        :type quote_asset: ERC20
        :param event: The event to convert.
        :type event: BaseEvent
        :param wallet: The wallet address associated with the event.
        :type wallet: ChecksumAddress
        :return: The created OrderEvent.
        :rtype: OrderEvent
        :raises Exception: If the event cannot be converted into an OrderEvent. This occurs if the Base event has a type
            other than EmitOfferEvent, EmitCancelEvent, EmitTakeEvent or EmitDeleteEvent
        """
        bid_identifier = cls._bid_identifier(
            base_asset=base_asset, quote_asset=quote_asset
        )

        if isinstance(event, EmitOfferEvent):
            if bid_identifier == event.pair:
                return cls._build_order(
                    event=event,
                    base_asset=base_asset,
                    quote_asset=quote_asset,
                    side=OrderSide.BUY,
                    order_type=OrderType.LIMIT,
                    base_amt=event.buy_amt,
                    quote_amt=event.pay_amt,
                )
            else:
                return cls._build_order(
                    event=event,
                    base_asset=base_asset,
                    quote_asset=quote_asset,
                    side=OrderSide.SELL,
                    order_type=OrderType.LIMIT,
                    base_amt=event.pay_amt,
                    quote_amt=event.buy_amt,
                )
        elif isinstance(event, EmitCancelEvent):
            if bid_identifier == event.pair:
                return cls._build_order(
                    event=event,
                    base_asset=base_asset,
                    quote_asset=quote_asset,
                    side=OrderSide.BUY,
                    order_type=OrderType.CANCEL,
                    base_amt=event.buy_amt,
                    quote_amt=event.pay_amt,
                )
            else:
                return cls._build_order(
                    event=event,
                    base_asset=base_asset,
                    quote_asset=quote_asset,
                    side=OrderSide.SELL,
                    order_type=OrderType.CANCEL,
                    base_amt=event.pay_amt,
                    quote_amt=event.buy_amt,
                )
        elif isinstance(event, EmitTakeEvent):
            # This is nuanced as we can either receive a take event for a market order or limit order we placed. When
            # we take the bid_identifier indicates selling into the bids while the ask identifier indicates buying
            # into the asks. While the reverse is true for when we make.
            if bid_identifier == event.pair:
                return cls._build_order(
                    event=event,
                    base_asset=base_asset,
                    quote_asset=quote_asset,
                    side=OrderSide.BUY if wallet == event.maker else OrderSide.SELL,
                    order_type=OrderType.LIMIT_TAKEN
                    if wallet == event.maker
                    else OrderType.MARKET,
                    base_amt=event.give_amt,
                    quote_amt=event.take_amt,
                )
            else:
                return cls._build_order(
                    event=event,
                    base_asset=base_asset,
                    quote_asset=quote_asset,
                    side=OrderSide.SELL if wallet == event.maker else OrderSide.BUY,
                    order_type=OrderType.LIMIT_TAKEN
                    if wallet == event.maker
                    else OrderType.MARKET,
                    base_amt=event.take_amt,
                    quote_amt=event.give_amt,
                )
        elif isinstance(event, EmitDeleteEvent):
            if bid_identifier == event.pair:
                return cls._build_order(
                    event=event,
                    base_asset=base_asset,
                    quote_asset=quote_asset,
                    side=None,
                    order_type=OrderType.LIMIT_DELETED,
                    base_amt=None,
                    quote_amt=None,
                )
            else:
                return cls._build_order(
                    event=event,
                    base_asset=base_asset,
                    quote_asset=quote_asset,
                    side=None,
                    order_type=OrderType.LIMIT_DELETED,
                    base_amt=None,
                    quote_amt=None,
                )

        else:
            Exception(f"{event.__class__} cannot be converted into an OrderEvent")

    @classmethod
    def _build_order(
        cls,
        event: Union[EmitOfferEvent, EmitCancelEvent, EmitTakeEvent, EmitDeleteEvent],
        base_asset: ERC20,
        quote_asset: ERC20,
        side: Optional[OrderSide],
        order_type: OrderType,
        base_amt: Optional[int],
        quote_amt: Optional[int],
    ) -> "OrderEvent":
        """Build an OrderEvent from event data.

        :param event: The event data.
        :type event: Union[EmitOfferEvent, EmitCancelEvent, EmitTakeEvent]
        :param base_asset: The base asset associated with the event.
        :type base_asset: ERC20
        :param quote_asset: The quote asset associated with the event.
        :type quote_asset: ERC20
        :param side: The order side.
        :type side: Optional[OrderSide]
        :param order_type: The order type.
        :type order_type: OrderType
        :param base_amt: The base amount of the order.
        :type base_amt: Optional[int]
        :param quote_amt: The quote amount of the order.
        :type quote_amt: Optional[int]
        :return: The constructed OrderEvent.
        :rtype: OrderEvent
        """
        size = base_asset.to_decimal(base_amt) if base_amt else None
        price = quote_asset.to_decimal(quote_amt) / size if quote_amt else None

        return cls(
            limit_order_id=event.id,
            limit_order_owner=event.maker,
            market_order_owner=event.taker
            if isinstance(event, EmitTakeEvent)
            else None,
            pair_name=f"{base_asset.symbol}/{quote_asset.symbol}",
            order_side=side,
            order_type=order_type,
            size=size,
            price=price,
        )

    @staticmethod
    def _bid_identifier(base_asset: ERC20, quote_asset: ERC20) -> str:
        return Web3.solidity_keccak(
            abi_types=["address", "address"],
            values=[
                quote_asset.address,
                base_asset.address,
            ],
        ).hex()

    def __repr__(self):
        items = ("{}={!r}".format(k, self.__dict__[k]) for k in self.__dict__)
        return "{}({})".format(type(self).__name__, ", ".join(items))


class FeeEvent:
    def __init__(
        self,
        id: int,
        pair_name: str,
        fee_to: ChecksumAddress,
        market_order_owner: ChecksumAddress,
        fee: Decimal,
        fee_asset: str,
    ):
        self.id = id
        self.pair_name = pair_name
        self.fee_to = fee_to
        self.market_order_owner = market_order_owner
        self.fee = fee
        self.fee_asset = fee_asset

    @classmethod
    def from_event(
        cls,
        pair_name: str,
        asset: ERC20,
        event: EmitFeeEvent,
    ) -> "FeeEvent":
        return cls(
            id=event.id,
            pair_name=pair_name,
            fee_to=event.fee_to,
            market_order_owner=event.taker,
            fee=asset.to_decimal(event.fee_amt),
            fee_asset=asset.symbol,
        )

    def __repr__(self):
        items = ("{}={!r}".format(k, self.__dict__[k]) for k in self.__dict__)
        return "{}({})".format(type(self).__name__, ", ".join(items))



import logging
from typing import List, Union

from hexbytes import HexBytes
from web3 import Web3
from web3.contract import Contract
from web3.logs import DISCARD
from web3.types import EventData, TxReceipt, TxParams

from rubi.contracts.contract_types import TransactionReceipt, BaseEvent

logger = logging.getLogger(__name__)


class TransactionHandler:
    """
    The transaction handler handles submitting transactions to chain and querying transaction receipts.

    :param w3: Web3 instance
    :type w3: Web3
    :param contracts: A list of contracts that will be used for decoding the logs on TxReceipts.
    :type contracts: List[Contract]
    """

    def __init__(self, w3: Web3, contracts: List[Contract]):
        self.w3 = w3
        self.contracts = contracts

    def add_contract(self, contract: Contract):
        """Add a contract to the list of contracts that are used to decode logs on TxReceipts.

        :param contract: A contracts to add to the contracts that will be used for decoding the logs on TxReceipts.
        :type contract: Contract
        """
        self.contracts.append(contract)

    def execute_transaction(
        self,
        transaction: TxParams,
        key: str,
    ) -> TransactionReceipt:
        """Execute a transaction by signing it with the given key and then submitting it to chain. Then wait for the
        transaction receipt for the transaction.

        :param transaction: The transaction to execute
        :type transaction: TxParams
        :param key: The private key to sign the transaction.
        :type key: str
        :return: The transaction receipt of the executed transaction.
        :rtype: TransactionReceipt
        """
        if "pair_names" in transaction:
            del transaction["pair_names"]

        signed_transaction = self.w3.eth.account.sign_transaction(
            transaction_dict=transaction, private_key=key
        )

        try:
            self.w3.eth.send_raw_transaction(signed_transaction.rawTransaction)
        except Exception as e:
            logger.error(f"Error trying to send transaction: {e}")
            raise e

        return self._wait_for_transaction_receipt(
            transaction_hash=signed_transaction.hash
        )

    def get_transaction_receipt(self, transaction_hash: str) -> TransactionReceipt:
        """Get a transaction receipt for the give transaction_hash.

        :param transaction_hash: The transaction hash.
        :type transaction_hash: str
        :return: A TransactionReceipt for the transaction hash.
        :rtype: TransactionReceipt
        """
        return self._wait_for_transaction_receipt(transaction_hash=transaction_hash)

    ######################################################################
    # helper methods
    ######################################################################

    def _wait_for_transaction_receipt(
        self,
        transaction_hash: str,
    ) -> TransactionReceipt:
        """Get the transaction receipt for the given transaction hash.

        :param transaction_hash: The transaction hash of the transaction.
        :type transaction_hash: str
        :return: The transaction receipt of the given transaction hash
        :rtype: TransactionReceipt
        """

        tx_receipt = self.w3.eth.wait_for_transaction_receipt(
            HexBytes(transaction_hash)
        )

        raw_events = self._process_receipt_logs_into_raw_events(receipt=tx_receipt)

        result = TransactionReceipt.from_tx_receipt(
            tx_receipt=tx_receipt, raw_events=raw_events
        )

        return result

    def _process_receipt_logs_into_raw_events(
        self, receipt: TxReceipt
    ) -> List[Union[BaseEvent, EventData]]:
        """
        Processes the logs of a given transaction receipt and returns a list of events associated with the transaction.

        :param receipt: The transaction receipt dictionary.
        :type receipt: TxReceipt
        :return: The list of events associated with the given transaction receipt
        :rtype: List[Union[BaseEvent, EventData]]
        """

        raw_events = []
        for contract in self.contracts:
            event_names = list(
                map(lambda event: event.event_name, contract.events)  # noqa
            )

            if len(event_names) == 0:
                continue

            for event_name in event_names:
                # TODO: Figure out why this breaks things for some reason
                if event_name == "LogNote":
                    continue

                transaction_events_data = contract.events[event_name]().process_receipt(
                    receipt, DISCARD
                )

                for event_data in transaction_events_data:
                    event = BaseEvent.from_raw(
                        name=event_name,
                        address=event_data["address"],
                        block_number=event_data["blockNumber"],
                        **event_data["args"],
                    )

                    if event:
                        raw_events.append(event)

        return raw_events



import logging
from _decimal import Decimal
from multiprocessing import Queue
from typing import Optional, Dict, List, Any, Type, Union

from eth_typing import ChecksumAddress
from web3.types import TxParams, EventData

from rubi import (
    Client,
    Network,
    LimitOrder,
    OrderEvent,
    OrderType,
    EmitTakeEvent,
    EmitDeleteEvent,
    TransactionReceipt,
    BaseEvent,
    EmitFeeEvent,
    FeeEvent,
    EmitCancelEvent,
    EmitOfferEvent,
)

logger = logging.getLogger(__name__)


class OrderTrackingClient(Client):
    """This class is a client for Rubicon that tracks orders on Rubicon for you.

    :param network: A Network instance
    :type network: Network
    :param pair_names: The list of pair names that the client is going to track.
    :type pair_names: List[str]
    :param message_queue: Optional message queue for processing events (optional, default is None).
    :type message_queue: Optional[Queue]
    :param wallet: Wallet address (optional, default is None).
    :type wallet: Optional[ChecksumAddress]
    :param key: Key for the wallet (optional, default is None).
    :type key: Optional[str]
    """

    def __init__(
        self,
        network: Network,
        pair_names: List[str],
        message_queue: Optional[Queue] = None,
        wallet: Optional[Union[ChecksumAddress, str]] = None,
        key: Optional[str] = None,
    ):
        super().__init__(
            network=network, message_queue=message_queue, wallet=wallet, key=key
        )

        limit_orders_from_subgraph = self.get_offers(
            maker=self.wallet, open=True, as_dataframe=False, pair_names=pair_names
        )

        self.open_limit_orders: Dict[int, LimitOrder] = {}
        if limit_orders_from_subgraph:
            for limit_order in limit_orders_from_subgraph:
                self.open_limit_orders[limit_order.order_id] = limit_order

        self.pairs_with_registered_event_listeners: List[str] = []
        self._register_listeners(pair_names=pair_names)

    @classmethod
    def from_http_node_url(
        cls,
        http_node_url: str,
        pair_names: List[str] = None,
        message_queue: Optional[Queue] = None,
        wallet: Optional[Union[ChecksumAddress, str]] = None,
        key: Optional[str] = None,
        custom_token_addresses_file: Optional[str] = None,
        **kwargs,
    ):
        """Initialize a Client using a http_node_url.

        :param http_node_url: URL of the HTTP node.
        :type http_node_url: str
        :param pair_names: The list of pair names that the client is going to track.
        :type pair_names: List[str]
        :param message_queue: Optional message queue for processing events (optional, default is None).
        :type message_queue: Optional[Queue]
        :param wallet: Wallet address (optional, default is None).
        :type wallet: Optional[Union[ChecksumAddress, str]]
        :param key: Key for the wallet (optional, default is None).
        :type key: str
        :param custom_token_addresses_file: The name of a yaml file (relative to the current working directory) with
            custom token addresses. Overwrites the token config found in network_config/{chain}/network.yaml.
            (optional, default is None).
        :type custom_token_addresses_file: Optional[str]
        """
        if pair_names is None:
            raise Exception(
                "OrderTrackingClient cannot be instantiated without a list of pair_names."
            )

        network = Network.from_http_node_url(
            http_node_url=http_node_url,
            custom_token_addresses_file=custom_token_addresses_file,
        )

        return cls(
            network=network,
            pair_names=pair_names,
            message_queue=message_queue,
            wallet=wallet,
            key=key,
        )

    ######################################################################
    # overridden methods
    ######################################################################

    def execute_transaction(self, transaction: TxParams) -> TransactionReceipt:
        """Execute the passed transaction.

        :param transaction: The transaction hash.
        :type transaction: TxParams
        :return: A TransactionReceipt of the executed transaction.
        :rtype: TransactionReceipt
        :return:
        """

        pair_names = transaction["pair_names"] if "pair_names" in transaction else None

        transaction_receipt = self.network.transaction_handler.execute_transaction(
            transaction=transaction, key=self._key
        )

        processed_transaction_receipt = self._handle_transaction_receipt_raw_events(
            transaction_receipt=transaction_receipt,
            pair_names=pair_names,
        )

        if pair_names:
            self._update_active_limit_orders(
                events=processed_transaction_receipt.events
            )

        return processed_transaction_receipt

    ######################################################################
    # order tracking methods
    ######################################################################

    def _default_event_handler(
        self, pair_name: str, event_type: Type[BaseEvent], event_data: EventData
    ) -> None:
        """The default event handler function used by the event poller. It processes the retrieved events
        and adds the corresponding order events to the message queue of the client.

        :param pair_name: Name of the pair associated with the event.
        :type pair_name: str
        :param event_type: Type of the event.
        :type event_type: Type[BaseEvent]
        :param event_data: Data of the retrieved event.
        :type event_data: EventData
        """
        raw_event = event_type(
            block_number=event_data["blockNumber"], **event_data["args"]
        )

        if raw_event.client_filter(wallet=self.wallet):
            if isinstance(raw_event, EmitFeeEvent):
                asset = self.network.tokens[raw_event.asset]

                event = FeeEvent.from_event(
                    pair_name=pair_name, asset=asset, event=raw_event
                )
            else:
                base_asset, quote_asset = pair_name.split("/")

                event = OrderEvent.from_event(
                    base_asset=self.network.tokens[base_asset],
                    quote_asset=self.network.tokens[quote_asset],
                    event=raw_event,
                    wallet=self.wallet,
                )

                self._update_active_limit_orders(events=[event])

            self.message_queue.put(event)

    def _update_active_limit_orders(self, events: List[Any]) -> None:
        """Update active limit order based on incoming events"""
        for event in events:
            if not isinstance(event, OrderEvent):
                continue

            event: OrderEvent
            if event.limit_order_owner != self.wallet:
                continue

            match event.order_type:
                case OrderType.LIMIT:
                    if self.open_limit_orders.get(event.limit_order_id):
                        # If we are already tracking the order then don't add it again
                        continue

                    self.open_limit_orders[
                        event.limit_order_id
                    ] = LimitOrder.from_order_event(order_event=event)
                case OrderType.LIMIT_TAKEN:
                    taken_order = self.open_limit_orders[event.limit_order_id]

                    if taken_order.remaining_size - event.size <= Decimal("0"):
                        self._delete_active_limit_order(order_id=event.limit_order_id)
                    else:
                        self.open_limit_orders[event.limit_order_id].update_with_take(
                            order_event=event
                        )
                case OrderType.CANCEL:
                    self._delete_active_limit_order(order_id=event.limit_order_id)
                case OrderType.LIMIT_DELETED:
                    self._delete_active_limit_order(order_id=event.limit_order_id)

    def _delete_active_limit_order(self, order_id: int) -> None:
        """Delete active limit order"""
        try:
            del self.open_limit_orders[order_id]
        except KeyError:
            logger.debug(
                f"Limit order {order_id} already removed from active limit orders."
            )

    def _register_listeners(self, pair_names: List[str]) -> None:
        """Register event listeners"""
        new_pair_names = [
            pair_name
            for pair_name in pair_names
            if pair_name not in self.pairs_with_registered_event_listeners
        ]

        self.pairs_with_registered_event_listeners.extend(new_pair_names)

        for pair_name in new_pair_names:
            self.start_event_poller(
                pair_name=pair_name,
                event_type=EmitOfferEvent,
                filters={"maker": self.wallet},
            )
            self.start_event_poller(
                pair_name=pair_name,
                event_type=EmitTakeEvent,
                filters={"maker": self.wallet},
            )
            self.start_event_poller(
                pair_name=pair_name,
                event_type=EmitCancelEvent,
                filters={"maker": self.wallet},
            )
            self.start_event_poller(
                pair_name=pair_name,
                event_type=EmitDeleteEvent,
                filters={"maker": self.wallet},
            )



import logging
from _decimal import Decimal
from multiprocessing import Queue
from threading import Thread
from time import sleep
from typing import Union, List, Optional, Dict, Type, Any, Callable, Tuple

import pandas as pd
from eth_typing import ChecksumAddress
from web3.types import EventData, Nonce, TxParams

from rubi import LimitOrder
from rubi.contracts import (
    ERC20,
    TransactionReceipt,
    EmitFeeEvent,
    EmitOfferEvent,
    EmitTakeEvent,
    EmitCancelEvent,
    EmitApproval,
    EmitTransfer,
)
from rubi.data import MarketData, SubgraphOffer
from rubi.network import (
    Network,
)
from rubi.rubicon_types import (
    OrderSide,
    NewMarketOrder,
    NewLimitOrder,
    OrderBook,
    BaseEvent,
    FeeEvent,
    OrderEvent,
    NewCancelOrder,
    UpdateLimitOrder,
    Approval,
    RubiconMarketApproval,
    RubiconRouterApproval,
    Transfer,
    ApprovalEvent,
    TransferEvent,
)

logger = logging.getLogger(__name__)


class Client:
    """This class is a client for Rubicon. It aims to provide a simple and understandable interface when interacting
    with the Rubicon protocol. If not instantiated with a wallet and key then all the methods that require signing
    will throw an error.

    :param network: A Network instance
    :type network: Network
    :param message_queue: Optional message queue for processing events (optional, default is None).
    :type message_queue: Optional[Queue]
    :param wallet: Wallet address (optional, default is None).
    :type wallet: Optional[ChecksumAddress]
    :param key: Key for the wallet (optional, default is None).
    :type key: Optional[str]
    """

    def __init__(
        self,
        network: Network,
        message_queue: Optional[Queue] = None,
        wallet: Optional[Union[ChecksumAddress, str]] = None,
        key: Optional[str] = None,
    ):
        """constructor method."""
        self.network = network

        self.message_queue = message_queue  # type: Queue | None

        # Authentication
        self.wallet = (
            self.network.w3.to_checksum_address(wallet) if wallet else wallet
        )  # type: ChecksumAddress |  None
        self._key = key  # type: str |  None

        self.market_data = MarketData.from_network(network=network)

    @classmethod
    def from_http_node_url(
        cls,
        http_node_url: str,
        message_queue: Optional[Queue] = None,
        wallet: Optional[Union[ChecksumAddress, str]] = None,
        key: Optional[str] = None,
        custom_token_addresses_file: Optional[str] = None,
        **kwargs,
    ):
        """Initialize a Client using a http_node_url.

        :param http_node_url: URL of the HTTP node.
        :type http_node_url: str
        :param message_queue: Optional message queue for processing events (optional, default is None).
        :type message_queue: Optional[Queue]
        :param wallet: Wallet address (optional, default is None).
        :type wallet: Optional[Union[ChecksumAddress, str]]
        :param key: Key for the wallet (optional, default is None).
        :type key: str
        :param custom_token_addresses_file: The name of a yaml file (relative to the current working directory) with
            custom token addresses. Overwrites the token config found in network_config/{chain}/network.yaml.
            (optional, default is None).
        :type custom_token_addresses_file: Optional[str]
        """
        network = Network.from_http_node_url(
            http_node_url=http_node_url,
            custom_token_addresses_file=custom_token_addresses_file,
        )

        return cls(
            network=network,
            message_queue=message_queue,
            wallet=wallet,
            key=key,
        )

    ######################################################################
    # transaction methods
    ######################################################################

    def get_nonce(self) -> Nonce:
        """Get the current transaction count of the wallet to determine the nonce

        :return: The current nonce of the wallet
        :rtype: Nonce
        """
        return self.network.w3.eth.get_transaction_count(self.wallet)

    def get_transaction_receipt(
        self,
        transaction_hash: str,
        pair_names: Optional[List[str]] = None,
    ) -> TransactionReceipt:
        """Get the transaction receipt for a given transaction hash.

        :param transaction_hash: The transaction hash.
        :type transaction_hash: str
        :param pair_names: If handling a transaction that interacted with the rubicon market we can decode events into
            human-readable format if we know the pairs in the transaction. (defaults to None and no decoding is done)
        :type pair_names: Optional[List[str]]
        :return: A TransactionReceipt for the transaction hash.
        :rtype: TransactionReceipt
        :return:
        """

        transaction_receipt = self.network.transaction_handler.get_transaction_receipt(
            transaction_hash=transaction_hash,
        )

        return self._handle_transaction_receipt_raw_events(
            transaction_receipt=transaction_receipt,
            pair_names=pair_names,
        )

    def execute_transaction(self, transaction: TxParams) -> TransactionReceipt:
        """Execute the passed transaction.

        :param transaction: The transaction hash.
        :type transaction: TxParams
        :return: A TransactionReceipt of the executed transaction.
        :rtype: TransactionReceipt
        :return:
        """
        pair_names = transaction["pair_names"] if "pair_names" in transaction else None

        transaction_receipt = self.network.transaction_handler.execute_transaction(
            transaction=transaction, key=self._key
        )

        processed_transaction_receipt = self._handle_transaction_receipt_raw_events(
            transaction_receipt=transaction_receipt,
            pair_names=pair_names,
        )

        return processed_transaction_receipt

    ######################################################################
    # token methods
    ######################################################################

    def get_balance(
        self, token: str, wallet: Optional[ChecksumAddress] = None
    ) -> Decimal:
        """Get balance of a token.

        :param token: The token to get the balance of
        :type token: str
        :param wallet: The wallet balance to check (Optional, defaults to client wallet)
        :type wallet: ChecksumAddress
        :return: The token balance of the wallet
        :rtype: Decimal
        """
        if not wallet:
            wallet = self.wallet

        balance = self.network.tokens[token].balance_of(account=wallet)

        return self.network.tokens[token].to_decimal(balance)

    def get_allowance(self, token: str, spender: ChecksumAddress) -> Decimal:
        """Get a spenders allowance for a certain token.

        :param token: The token to check the allowance of
        :type token: str
        :param spender: The spender address
        :type spender: ChecksumAddress
        :return: The allowance of the spender
        :rtype: Decimal
        """

        allowance = self.network.tokens[token].allowance(
            owner=self.wallet, spender=spender
        )

        return self.network.tokens[token].to_decimal(allowance)

    # TODO: revisit as the safer thing is to set approval to 0 and then set approval to new_allowance
    #  or use increaseAllowance and decreaseAllowance but the current abi does not support these methods
    #  See: https://github.com/ethereum/EIPs/issues/20#issuecomment-263524729
    def approve(
        self,
        approval: Approval,
        nonce: Optional[int] = None,
        gas: Optional[int] = None,
        max_fee_per_gas: Optional[int] = None,
        max_priority_fee_per_gas: Optional[int] = None,
    ) -> TxParams:
        """
        Construct an approval transaction.

        :param approval: The approval of a spender of an ERC20
        :type approval: Approval | RubiconMarketApproval | RubiconRouterApproval
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce.
            (optional, default is None)
        :type nonce: Optional[int]
        :param gas: gas limit for the transaction. If None is passed then w3.eth.estimate_gas is used.
        :type gas: Optional[int]
        :param max_fee_per_gas: max fee that can be paid for gas, defaults to
            max_priority_fee (from chain) + (2 * base fee per gas of latest block) (optional, default is None)
        :type max_fee_per_gas: Optional[int]
        :param max_priority_fee_per_gas: max priority fee that can be paid for gas, defaults to calling the chain to
            estimate the max_priority_fee_per_gas (optional, default is None)
        :type max_priority_fee_per_gas: Optional[int]
        :return: The built transaction.
        :rtype: TxParams
        """
        amount = self.network.tokens[approval.token].to_integer(approval.amount)
        spender = approval.spender

        if isinstance(approval, RubiconMarketApproval):
            spender = self.network.rubicon_market.address
        elif isinstance(approval, RubiconRouterApproval):
            spender = self.network.rubicon_router.address
        elif spender is None:
            raise Exception("A spender must be provided for an approval")

        return self.network.tokens[approval.token].approve(
            spender=spender,
            amount=amount,
            wallet=self.wallet,
            nonce=nonce,
            gas=gas,
            max_fee_per_gas=max_fee_per_gas,
            max_priority_fee_per_gas=max_priority_fee_per_gas,
        )

    def transfer(
        self,
        transfer: Transfer,
        nonce: Optional[int] = None,
        gas: Optional[int] = None,
        max_fee_per_gas: Optional[int] = None,
        max_priority_fee_per_gas: Optional[int] = None,
    ) -> TxParams:
        """
        Construct a transfer transaction.

        :param transfer: The transfer of an ERC20 to a certain address
        :type transfer: Transfer
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce.
            (optional, default is None)
        :type nonce: Optional[int]
        :param gas: gas limit for the transaction. If None is passed then w3.eth.estimate_gas is used.
        :type gas: Optional[int]
        :param max_fee_per_gas: max fee that can be paid for gas, defaults to
            max_priority_fee (from chain) + (2 * base fee per gas of latest block) (optional, default is None)
        :type max_fee_per_gas: Optional[int]
        :param max_priority_fee_per_gas: max priority fee that can be paid for gas, defaults to calling the chain to
            estimate the max_priority_fee_per_gas (optional, default is None)
        :type max_priority_fee_per_gas: Optional[int]
        :return: The built transaction.
        :rtype: TxParams
        """
        amount = self.network.tokens[transfer.token].to_integer(transfer.amount)

        return self.network.tokens[transfer.token].transfer(
            recipient=transfer.recipient,
            amount=amount,
            wallet=self.wallet,
            nonce=nonce,
            gas=gas,
            max_fee_per_gas=max_fee_per_gas,
            max_priority_fee_per_gas=max_priority_fee_per_gas,
        )

    ######################################################################
    # orderbook methods
    ######################################################################

    def get_orderbook(self, pair_name: str) -> OrderBook:
        """Retrieve the order book for a specific pair from the Rubicon Router.

        :param pair_name: Name of the pair to retrieve the order book for.
        :type pair_name: str
        :return: The order book for the specified pair.
        :rtype: OrderBook
        :raises PairDoesNotExistException: If the pair does not exist in the client.
        """
        base_asset, quote_asset = pair_name.split("/")

        rubicon_offer_book = self.network.rubicon_router.get_book_from_pair(
            asset=self.network.tokens[base_asset].address,
            quote=self.network.tokens[quote_asset].address,
        )

        return OrderBook.from_rubicon_offer_book(
            offer_book=rubicon_offer_book,
            base_asset=self.network.tokens[base_asset],
            quote_asset=self.network.tokens[quote_asset],
        )

    def start_orderbook_poller(self, pair_name: str, poll_time: int = 2) -> None:
        """Starts a background thread that continuously polls the order book for the specified pair
        at a specified polling interval. The retrieved order book is added to the message queue of the client.
        The poller will run until the pair is removed from the client.

        :param pair_name: Name of the pair to start the order book poller for.
        :type pair_name: str
        :param poll_time: Polling interval in seconds, defaults to 2 seconds.
        :type poll_time: int, optional
        :raises Exception: If the message queue is not configured.
        :raises PairDoesNotExistException: If the pair does not exist in the client.
        """

        if self.message_queue is None:
            raise Exception(
                "Orderbook poller is configured to place messages on the message queue. Message queue cannot be none"
            )

        base_asset, quote_asset = pair_name.split("/")
        if (
            base_asset not in self.network.tokens.keys()
            or quote_asset not in self.network.tokens.keys()
        ):
            raise Exception(
                f"Cannot start orderbook poller for {base_asset}/{quote_asset} as these assets are not in"
                f"the clients token set: {self.network.tokens.keys()}"
            )

        thread = Thread(
            target=self._start_orderbook_poller,
            kwargs={"pair_name": pair_name, "poll_time": poll_time},
            daemon=True,
        )
        thread.start()

    # TODO: look at using a listener instead of poller (we will probably need to listen to events)
    def _start_orderbook_poller(self, pair_name: str, poll_time: int = 2) -> None:
        """The internal implementation of the order book poller. It continuously retrieves the order book for the
        specified pair and adds it to the message queue of the client.

        :param pair_name: Name of the pair to start the order book poller for.
        :type pair_name: str
        :param poll_time: Polling interval in seconds, defaults to 2 seconds.
        :type poll_time: int, optional
        """
        polling: bool = True
        while polling:
            try:
                self.message_queue.put(self.get_orderbook(pair_name=pair_name))
            except Exception as e:
                logger.error(e)
            sleep(poll_time)

    ######################################################################
    # event methods
    ######################################################################

    def start_event_poller(
        self,
        pair_name: str,
        event_type: Type[BaseEvent],
        filters: Optional[Dict[str, Any]] = None,
        event_handler: Optional[Callable] = None,
        poll_time: int = 2,
    ) -> None:
        """Starts a background event poller that continuously listens for events of the specified event type
        related to the specified pair. The retrieved events are processed by the event handler and added to the message
        queue of the client. The poller will run until the pair is removed from the client.

        :param pair_name: Name of the pair to start the event poller for.
        :type pair_name: str
        :param event_type: Type of the event to listen for.
        :type event_type: Type[BaseEvent]
        :param filters: Optional filters to apply when retrieving events, defaults to the events default filters
            (optional, default is None). These are added to the default filters for the event
        :type filters: Optional[Dict[str, Any]], optional
        :param event_handler: Optional event handler function to process the retrieved events, defaults to the
            self._default_event_handler (optional, default is None).
        :type event_handler: Optional[Callable], optional
        :param poll_time: Polling interval in seconds, defaults to 2 seconds.
        :type poll_time: int, optional
        :raises Exception: If the message queue is not configured.
        :raises PairDoesNotExistException: If the pair does not exist in the client.
        """
        if self.message_queue is None:
            raise Exception(
                "Event poller is configured to place messages on the message queue. Message queue"
                "cannot be none."
            )

        base_asset, quote_asset = pair_name.split("/")

        bid_identifier = self.network.w3.solidity_keccak(
            abi_types=["address", "address"],
            values=[
                self.network.tokens[quote_asset].address,
                self.network.tokens[base_asset].address,
            ],
        ).hex()
        ask_identifier = self.network.w3.solidity_keccak(
            abi_types=["address", "address"],
            values=[
                self.network.tokens[base_asset].address,
                self.network.tokens[quote_asset].address,
            ],
        ).hex()

        argument_filters = event_type.default_filters(
            bid_identifier=bid_identifier, ask_identifier=ask_identifier
        )

        if filters is not None:
            # TODO: add check that filters are valid, if i remember correctly i think we can only filter on indexed
            #  params. i bet there is a function to check this
            argument_filters.update(filters)

        event_type.get_event_contract(
            market=self.network.rubicon_market, router=self.network.rubicon_router
        ).start_event_poller(
            pair_name=pair_name,
            event_type=event_type,
            argument_filters=argument_filters,
            event_handler=self._default_event_handler
            if event_handler is None
            else event_handler,
            poll_time=poll_time,
        )

    def _default_event_handler(
        self, pair_name: str, event_type: Type[BaseEvent], event_data: EventData
    ) -> None:
        """The default event handler function used by the event poller. It processes the retrieved events
        and adds the corresponding order events to the message queue of the client.

        :param pair_name: Name of the pair associated with the event.
        :type pair_name: str
        :param event_type: Type of the event.
        :type event_type: Type[BaseEvent]
        :param event_data: Data of the retrieved event.
        :type event_data: EventData
        """
        raw_event = event_type(
            block_number=event_data["blockNumber"], **event_data["args"]
        )

        if raw_event.client_filter(wallet=self.wallet):
            if isinstance(raw_event, EmitFeeEvent):
                asset = self.network.tokens[raw_event.asset]

                event = FeeEvent.from_event(
                    pair_name=pair_name, asset=asset, event=raw_event
                )
            else:
                base_asset, quote_asset = pair_name.split("/")

                event = OrderEvent.from_event(
                    base_asset=self.network.tokens[base_asset],
                    quote_asset=self.network.tokens[quote_asset],
                    event=raw_event,
                    wallet=self.wallet,
                )

            self.message_queue.put(event)

    ######################################################################
    # order methods
    ######################################################################

    def market_order(
        self,
        order: NewMarketOrder,
        nonce: Optional[int] = None,
        gas: Optional[int] = None,
        max_fee_per_gas: Optional[int] = None,
        max_priority_fee_per_gas: Optional[int] = None,
    ) -> TxParams:
        """Construct a new market order transaction. The corresponding market buy or sell method is called based on the
        order side.

        :param order: The market order to place.
        :type order: NewMarketOrder
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce.
            (optional, default is None)
        :type nonce: Optional[int]
        :param gas: gas limit for the transaction. If None is passed then w3.eth.estimate_gas is used.
        :type gas: Optional[int]
        :param max_fee_per_gas: max fee that can be paid for gas, defaults to
            max_priority_fee (from chain) + (2 * base fee per gas of latest block) (optional, default is None)
        :type max_fee_per_gas: Optional[int]
        :param max_priority_fee_per_gas: max priority fee that can be paid for gas, defaults to calling the chain to
            estimate the max_priority_fee_per_gas (optional, default is None)
        :type max_priority_fee_per_gas: Optional[int]
        :return: The transaction to execute the market order.
        :rtype: TxParams
        """
        base_asset, quote_asset = order.pair_name.split("/")

        match order.order_side:
            case OrderSide.BUY:
                transaction = self.network.rubicon_market.buy_all_amount(
                    buy_gem=self.network.tokens[base_asset].address,
                    buy_amt=self.network.tokens[base_asset].to_integer(order.size),
                    pay_gem=self.network.tokens[quote_asset].address,
                    max_fill_amount=self.network.tokens[quote_asset].to_integer(
                        order.worst_execution_price * order.size
                    ),
                    wallet=self.wallet,
                    nonce=nonce,
                    gas=gas,
                    max_fee_per_gas=max_fee_per_gas,
                    max_priority_fee_per_gas=max_priority_fee_per_gas,
                )
            case OrderSide.SELL:
                transaction = self.network.rubicon_market.sell_all_amount(
                    pay_gem=self.network.tokens[base_asset].address,
                    pay_amt=self.network.tokens[base_asset].to_integer(order.size),
                    buy_gem=self.network.tokens[quote_asset].address,
                    min_fill_amount=self.network.tokens[quote_asset].to_integer(
                        order.worst_execution_price * order.size
                    ),
                    wallet=self.wallet,
                    nonce=nonce,
                    gas=gas,
                    max_fee_per_gas=max_fee_per_gas,
                    max_priority_fee_per_gas=max_priority_fee_per_gas,
                )
            case _:
                raise Exception("OrderSide must be BUY or SELL")

        transaction["pair_names"] = [order.pair_name]

        return transaction

    def limit_order(
        self,
        order: NewLimitOrder,
        nonce: Optional[int] = None,
        gas: Optional[int] = None,
        max_fee_per_gas: Optional[int] = None,
        max_priority_fee_per_gas: Optional[int] = None,
    ) -> TxParams:
        """Construct a new limit order transaction.

        :param order: The limit order to place.
        :type order: NewLimitOrder
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce.
            (optional, default is None)
        :type nonce: Optional[int]
        :param gas: gas limit for the transaction. If None is passed then w3.eth.estimate_gas is used.
        :type gas: Optional[int]
        :param max_fee_per_gas: max fee that can be paid for gas, defaults to
            max_priority_fee (from chain) + (2 * base fee per gas of latest block) (optional, default is None)
        :type max_fee_per_gas: Optional[int]
        :param max_priority_fee_per_gas: max priority fee that can be paid for gas, defaults to calling the chain to
            estimate the max_priority_fee_per_gas (optional, default is None)
        :type max_priority_fee_per_gas: Optional[int]
        :return: The transaction to execute the limit order.
        :rtype: TxParams
        """
        base_asset, quote_asset = order.pair_name.split("/")

        match order.order_side:
            case OrderSide.BUY:
                transaction = self.network.rubicon_market.offer(
                    pay_amt=self.network.tokens[quote_asset].to_integer(
                        order.price * order.size
                    ),
                    pay_gem=self.network.tokens[quote_asset].address,
                    buy_amt=self.network.tokens[base_asset].to_integer(order.size),
                    buy_gem=self.network.tokens[base_asset].address,
                    wallet=self.wallet,
                    nonce=nonce,
                    gas=gas,
                    max_fee_per_gas=max_fee_per_gas,
                    max_priority_fee_per_gas=max_priority_fee_per_gas,
                )
            case OrderSide.SELL:
                transaction = self.network.rubicon_market.offer(
                    pay_amt=self.network.tokens[base_asset].to_integer(order.size),
                    pay_gem=self.network.tokens[base_asset].address,
                    buy_amt=self.network.tokens[quote_asset].to_integer(
                        order.price * order.size
                    ),
                    buy_gem=self.network.tokens[quote_asset].address,
                    wallet=self.wallet,
                    nonce=nonce,
                    gas=gas,
                    max_fee_per_gas=max_fee_per_gas,
                    max_priority_fee_per_gas=max_priority_fee_per_gas,
                )
            case _:
                raise Exception("OrderSide must be BUY or SELL")

        transaction["pair_names"] = [order.pair_name]

        return transaction

    def cancel_limit_order(
        self,
        order: NewCancelOrder,
        nonce: Optional[int] = None,
        gas: Optional[int] = None,
        max_fee_per_gas: Optional[int] = None,
        max_priority_fee_per_gas: Optional[int] = None,
    ) -> TxParams:
        """Construct a new limit order cancellation.

        :param order: The cancel order object.
        :type order: NewCancelOrder
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce.
            (optional, default is None)
        :type nonce: Optional[int]
        :param gas: gas limit for the transaction. If None is passed then w3.eth.estimate_gas is used.
        :type gas: Optional[int]
        :param max_fee_per_gas: max fee that can be paid for gas, defaults to
            max_priority_fee (from chain) + (2 * base fee per gas of latest block) (optional, default is None)
        :type max_fee_per_gas: Optional[int]
        :param max_priority_fee_per_gas: max priority fee that can be paid for gas, defaults to calling the chain to
            estimate the max_priority_fee_per_gas (optional, default is None)
        :type max_priority_fee_per_gas: Optional[int]
        :return: The transaction to execute the limit order cancellation.
        :rtype: TxParams
        """
        transaction = self.network.rubicon_market.cancel(
            id=order.order_id,
            wallet=self.wallet,
            nonce=nonce,
            gas=gas,
            max_fee_per_gas=max_fee_per_gas,
            max_priority_fee_per_gas=max_priority_fee_per_gas,
        )

        transaction["pair_names"] = [order.pair_name]

        return transaction

    def batch_limit_orders(
        self,
        orders: List[NewLimitOrder],
        nonce: Optional[int] = None,
        gas: Optional[int] = None,
        max_fee_per_gas: Optional[int] = None,
        max_priority_fee_per_gas: Optional[int] = None,
    ) -> TxParams:
        """Construct a transaction to place multiple limit orders in a batch.

        :param orders: The new limits orders to place.
        :type orders: List[NewLimitOrder]
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce.
            (optional, default is None)
        :type nonce: Optional[int]
        :param gas: gas limit for the transaction. If None is passed then w3.eth.estimate_gas is used.
        :type gas: Optional[int]
        :param max_fee_per_gas: max fee that can be paid for gas, defaults to
            max_priority_fee (from chain) + (2 * base fee per gas of latest block) (optional, default is None)
        :type max_fee_per_gas: Optional[int]
        :param max_priority_fee_per_gas: max priority fee that can be paid for gas, defaults to calling the chain to
            estimate the max_priority_fee_per_gas (optional, default is None)
        :type max_priority_fee_per_gas: Optional[int]
        :return: Transaction to execute the limit order batch.
        :rtype: TxParams
        """
        pay_amts = []
        pay_gems = []
        buy_amts = []
        buy_gems = []

        pair_names: List[str] = []
        for order in orders:
            base_asset, quote_asset = order.pair_name.split("/")
            pair_names.append(order.pair_name)

            match order.order_side:
                case OrderSide.BUY:
                    pay_amts.append(
                        self.network.tokens[quote_asset].to_integer(
                            order.price * order.size
                        )
                    )
                    pay_gems.append(self.network.tokens[quote_asset].address)
                    buy_amts.append(
                        self.network.tokens[base_asset].to_integer(order.size)
                    )
                    buy_gems.append(self.network.tokens[base_asset].address)
                case OrderSide.SELL:
                    pay_amts.append(
                        self.network.tokens[base_asset].to_integer(order.size)
                    )
                    pay_gems.append(self.network.tokens[base_asset].address)
                    buy_amts.append(
                        self.network.tokens[quote_asset].to_integer(
                            order.price * order.size
                        )
                    )
                    buy_gems.append(self.network.tokens[quote_asset].address)

        transaction = self.network.rubicon_market.batch_offer(
            pay_amts=pay_amts,
            pay_gems=pay_gems,
            buy_amts=buy_amts,
            buy_gems=buy_gems,
            wallet=self.wallet,
            nonce=nonce,
            gas=gas,
            max_fee_per_gas=max_fee_per_gas,
            max_priority_fee_per_gas=max_priority_fee_per_gas,
        )
        if transaction:
            transaction["pair_names"] = pair_names

        return transaction

    def batch_update_limit_orders(
        self,
        orders: List[UpdateLimitOrder],
        nonce: Optional[int] = None,
        gas: Optional[int] = None,
        max_fee_per_gas: Optional[int] = None,
        max_priority_fee_per_gas: Optional[int] = None,
    ) -> TxParams:
        """Construct a transaction to update multiple limit orders in a batch.

        :param orders: The new limits orders to place.
        :type orders: List[NewLimitOrder]
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce.
            (optional, default is None)
        :type nonce: Optional[int]
        :param gas: gas limit for the transaction. If None is passed then w3.eth.estimate_gas is used.
        :type gas: Optional[int]
        :param max_fee_per_gas: max fee that can be paid for gas, defaults to
            max_priority_fee (from chain) + (2 * base fee per gas of latest block) (optional, default is None)
        :type max_fee_per_gas: Optional[int]
        :param max_priority_fee_per_gas: max priority fee that can be paid for gas, defaults to calling the chain to
            estimate the max_priority_fee_per_gas (optional, default is None)
        :type max_priority_fee_per_gas: Optional[int]
        :return: The transaction to execute the update limit order batch.
        :rtype: TxParams
        """
        order_ids = []
        pay_amts = []
        pay_gems = []
        buy_amts = []
        buy_gems = []

        pair_names: List[str] = []
        for order in orders:
            base_asset, quote_asset = order.pair_name.split("/")
            pair_names.append(order.pair_name)

            order_ids.append(order.order_id)

            match order.order_side:
                case OrderSide.BUY:
                    pay_amts.append(
                        self.network.tokens[quote_asset].to_integer(
                            order.price * order.size
                        )
                    )
                    pay_gems.append(self.network.tokens[quote_asset].address)
                    buy_amts.append(
                        self.network.tokens[base_asset].to_integer(order.size)
                    )
                    buy_gems.append(self.network.tokens[base_asset].address)
                case OrderSide.SELL:
                    pay_amts.append(
                        self.network.tokens[base_asset].to_integer(order.size)
                    )
                    pay_gems.append(self.network.tokens[base_asset].address)
                    buy_amts.append(
                        self.network.tokens[quote_asset].to_integer(
                            order.price * order.size
                        )
                    )
                    buy_gems.append(self.network.tokens[quote_asset].address)

        transaction = self.network.rubicon_market.batch_requote(
            ids=order_ids,
            pay_amts=pay_amts,
            pay_gems=pay_gems,
            buy_amts=buy_amts,
            buy_gems=buy_gems,
            wallet=self.wallet,
            nonce=nonce,
            gas=gas,
            max_fee_per_gas=max_fee_per_gas,
            max_priority_fee_per_gas=max_priority_fee_per_gas,
        )

        transaction["pair_names"] = pair_names

        return transaction

    def batch_cancel_limit_orders(
        self,
        orders: List[NewCancelOrder],
        nonce: Optional[int] = None,
        gas: Optional[int] = None,
        max_fee_per_gas: Optional[int] = None,
        max_priority_fee_per_gas: Optional[int] = None,
    ) -> TxParams:
        """Construct a transaction to cancel multiple limit orders in a batch.

        :param orders: List of cancel limit orders
        :type orders: List[NewCancelOrder]
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce.
            (optional, default is None)
        :type nonce: Optional[int]
        :param gas: gas limit for the transaction. If None is passed then w3.eth.estimate_gas is used.
        :type gas: Optional[int]
        :param max_fee_per_gas: max fee that can be paid for gas, defaults to
            max_priority_fee (from chain) + (2 * base fee per gas of latest block) (optional, default is None)
        :type max_fee_per_gas: Optional[int]
        :param max_priority_fee_per_gas: max priority fee that can be paid for gas, defaults to calling the chain to
            estimate the max_priority_fee_per_gas (optional, default is None)
        :type max_priority_fee_per_gas: Optional[int]
        :return: The transaction to execute the cancel limit orders batch.
        :rtype: TxParams
        """
        order_ids = []

        pair_names: List[str] = []
        for order in orders:
            pair_names.append(order.pair_name)

            order_ids.append(order.order_id)

        transaction = self.network.rubicon_market.batch_cancel(
            ids=order_ids,
            wallet=self.wallet,
            nonce=nonce,
            gas=gas,
            max_fee_per_gas=max_fee_per_gas,
            max_priority_fee_per_gas=max_priority_fee_per_gas,
        )

        transaction["pair_names"] = pair_names

        return transaction

    ######################################################################
    # data methods
    ######################################################################

    def get_offers(
        self,
        maker: Optional[Union[ChecksumAddress, str]] = None,
        from_address: Optional[Union[ChecksumAddress, str]] = None,
        pair_names: Optional[List[str]] = None,
        book_side: Optional[
            OrderSide
        ] = None,  # TODO: decide if we should default to neutral
        open: Optional[bool] = None,
        start_time: Optional[int] = None,
        end_time: Optional[int] = None,
        start_block: Optional[int] = None,
        end_block: Optional[int] = None,
        first: int = 10000000,
        order_by: str = "timestamp",
        order_direction: str = "desc",
        as_dataframe: bool = True,
    ) -> Optional[pd.DataFrame] | List[LimitOrder]:
        # TODO: add support for multiple pair_names
        if len(pair_names) == 1:
            base_asset, quote_asset = pair_names[0].split("/")

            match book_side:
                case OrderSide.BUY:
                    result = self.market_data.get_offers(
                        maker=maker,
                        from_address=from_address,
                        buy_gem=self.network.tokens[base_asset].address,
                        pay_gem=self.network.tokens[quote_asset].address,
                        side=book_side.value.lower(),
                        open=open,
                        start_time=start_time,
                        end_time=end_time,
                        start_block=start_block,
                        end_block=end_block,
                        first=first,
                        order_by=order_by,
                        order_direction=order_direction,
                        as_dataframe=as_dataframe,
                    )
                case OrderSide.SELL:
                    result = self.market_data.get_offers(
                        maker=maker,
                        from_address=from_address,
                        buy_gem=self.network.tokens[quote_asset].address,
                        pay_gem=self.network.tokens[base_asset].address,
                        side=book_side.value.lower(),
                        open=open,
                        start_time=start_time,
                        end_time=end_time,
                        start_block=start_block,
                        end_block=end_block,
                        first=first,
                        order_by=order_by,
                        order_direction=order_direction,
                        as_dataframe=as_dataframe,
                    )
                case _:
                    bids = self.market_data.get_offers(
                        maker=maker,
                        from_address=from_address,
                        buy_gem=self.network.tokens[base_asset].address,
                        pay_gem=self.network.tokens[quote_asset].address,
                        side=OrderSide.BUY.value.lower(),
                        open=open,
                        start_time=start_time,
                        end_time=end_time,
                        start_block=start_block,
                        end_block=end_block,
                        first=first,
                        order_by=order_by,
                        order_direction=order_direction,
                        as_dataframe=as_dataframe,
                    )
                    asks = self.market_data.get_offers(
                        maker=maker,
                        from_address=from_address,
                        buy_gem=self.network.tokens[quote_asset].address,
                        pay_gem=self.network.tokens[base_asset].address,
                        side=OrderSide.SELL.value.lower(),
                        open=open,
                        start_time=start_time,
                        end_time=end_time,
                        start_block=start_block,
                        end_block=end_block,
                        first=first,
                        order_by=order_by,
                        order_direction=order_direction,
                        as_dataframe=as_dataframe,
                    )

                    if as_dataframe:
                        result = pd.concat([bids, asks]).reset_index(drop=True)
                    else:
                        result = bids + asks

        else:
            result = self.market_data.get_offers(
                maker=maker,
                from_address=from_address,
                buy_gem=None,
                pay_gem=None,
                side=None,
                open=open,
                start_time=start_time,
                end_time=end_time,
                start_block=start_block,
                end_block=end_block,
                first=first,
                order_by=order_by,
                order_direction=order_direction,
                as_dataframe=as_dataframe,
            )

        if isinstance(result, List):
            limit_orders: List[LimitOrder] = []

            for offer in result:
                base_asset, quote_asset = self._get_base_and_quote_asset(
                    raw=offer, pair_names=pair_names
                )
                if not base_asset or not quote_asset:
                    continue

                limit_orders.append(
                    LimitOrder.from_subgraph_offer(
                        base_asset=base_asset, quote_asset=quote_asset, offer=offer
                    )
                )
            return limit_orders

        return result

    def get_trades(
        self,
        taker: Optional[Union[ChecksumAddress, str]] = None,
        from_address: Optional[Union[ChecksumAddress, str]] = None,
        pair_names: Optional[List[str]] = None,
        book_side: Optional[OrderSide] = None,
        start_time: Optional[int] = None,
        end_time: Optional[int] = None,
        start_block: Optional[int] = None,
        end_block: Optional[int] = None,
        maker: Optional[
            Union[ChecksumAddress, str]
        ] = None,  # TODO: implement this with nested filtering
        maker_from_address: Optional[
            Union[ChecksumAddress, str]
        ] = None,  # TODO: implement this with nested filtering
        first: int = 10000000,
        order_by: str = "timestamp",
        order_direction: str = "desc",
    ) -> pd.DataFrame:
        if len(pair_names) == 1:
            base_asset, quote_asset = pair_names[0].split("/")

            match book_side:
                case OrderSide.BUY:
                    return self.market_data.get_trades(
                        taker=taker,
                        from_address=from_address,
                        take_gem=self.network.tokens[base_asset].address,
                        give_gem=self.network.tokens[quote_asset].address,
                        side=book_side.value.lower(),
                        start_time=start_time,
                        end_time=end_time,
                        start_block=start_block,
                        end_block=end_block,
                        maker=maker,
                        maker_from_address=maker_from_address,
                        first=first,
                        order_by=order_by,
                        order_direction=order_direction,
                    )
                case OrderSide.SELL:
                    return self.market_data.get_trades(
                        taker=taker,
                        from_address=from_address,
                        take_gem=self.network.tokens[quote_asset].address,
                        give_gem=self.network.tokens[base_asset].address,
                        side=book_side.value.lower(),
                        start_time=start_time,
                        end_time=end_time,
                        start_block=start_block,
                        end_block=end_block,
                        maker=maker,
                        maker_from_address=maker_from_address,
                        first=first,
                        order_by=order_by,
                        order_direction=order_direction,
                    )
                case _:
                    buys = self.market_data.get_trades(
                        taker=taker,
                        from_address=from_address,
                        take_gem=self.network.tokens[base_asset].address,
                        give_gem=self.network.tokens[quote_asset].address,
                        side=OrderSide.BUY.value.lower(),
                        start_time=start_time,
                        end_time=end_time,
                        start_block=start_block,
                        end_block=end_block,
                        maker=maker,
                        maker_from_address=maker_from_address,
                        first=first,
                        order_by=order_by,
                        order_direction=order_direction,
                    )
                    sells = self.market_data.get_trades(
                        taker=taker,
                        from_address=from_address,
                        take_gem=self.network.tokens[quote_asset].address,
                        give_gem=self.network.tokens[base_asset].address,
                        side=OrderSide.SELL.value.lower(),
                        start_time=start_time,
                        end_time=end_time,
                        start_block=start_block,
                        end_block=end_block,
                        maker=maker,
                        maker_from_address=maker_from_address,
                        first=first,
                        order_by=order_by,
                        order_direction=order_direction,
                    )

                    return pd.concat([buys, sells]).reset_index(drop=True)

        else:
            return self.market_data.get_trades(
                taker=taker,
                from_address=from_address,
                take_gem=None,
                give_gem=None,
                side=None,
                start_time=start_time,
                end_time=end_time,
                start_block=start_block,
                end_block=end_block,
                maker=maker,
                maker_from_address=maker_from_address,
                first=first,
                order_by=order_by,
                order_direction=order_direction,
            )

    ######################################################################
    # helper methods
    ######################################################################

    def _handle_transaction_receipt_raw_events(
        self,
        transaction_receipt: TransactionReceipt,
        pair_names: Optional[List[str]] = None,
    ) -> TransactionReceipt:
        """
        Transforms the raw transaction receipt events to human-readable events

        :param transaction_receipt:
        :type transaction_receipt: TransactionReceipt
        :return: The transaction receipt with human-readable events populated
        :rtype: TransactionReceipt
        """
        events: List[Any] = []
        for raw_event in transaction_receipt.raw_events:
            if (
                isinstance(raw_event, EmitOfferEvent)
                or isinstance(raw_event, EmitTakeEvent)
                or isinstance(raw_event, EmitCancelEvent)
            ):
                raw_event: Union[EmitOfferEvent, EmitTakeEvent, EmitCancelEvent]

                base_asset, quote_asset = self._get_base_and_quote_asset(
                    raw=raw_event, pair_names=pair_names
                )
                if not base_asset or not quote_asset:
                    continue

                events.append(
                    OrderEvent.from_event(
                        base_asset=base_asset,
                        quote_asset=quote_asset,
                        event=raw_event,
                        wallet=self.wallet,
                    )
                )
            if isinstance(raw_event, EmitApproval):
                try:
                    token = self.network.tokens[raw_event.address]
                except KeyError:
                    continue
                if raw_event.src == self.wallet or raw_event.guy == self.wallet:
                    events.append(
                        ApprovalEvent(
                            token=token.symbol,
                            amount=token.to_decimal(raw_event.wad),
                            spender=raw_event.guy,
                            source=raw_event.src,
                        )
                    )
            if isinstance(raw_event, EmitTransfer):
                try:
                    token = self.network.tokens[raw_event.address]
                except KeyError:
                    continue
                if raw_event.src == self.wallet or raw_event.dst == self.wallet:
                    events.append(
                        TransferEvent(
                            token=token.symbol,
                            amount=token.to_decimal(raw_event.wad),
                            recipient=raw_event.dst,
                            source=raw_event.src,
                        )
                    )
        # TODO: handle other events nicely

        transaction_receipt.set_events(events=events)

        return transaction_receipt

    def _get_base_and_quote_asset(
        self,
        raw: EmitOfferEvent | EmitTakeEvent | EmitCancelEvent | SubgraphOffer,
        pair_names: Optional[List[str]] = None,
    ) -> Tuple[Optional[ERC20], Optional[ERC20]]:
        """Get the base and quote asset of an event"""
        if pair_names is None:
            return None, None

        pay_gem = self.network.tokens[raw.pay_gem]
        buy_gem = self.network.tokens[raw.buy_gem]

        for pair_name in pair_names:
            base_asset, quote_asset = pair_name.split("/")

            if base_asset == pay_gem.symbol and quote_asset == buy_gem.symbol:
                return pay_gem, buy_gem
            elif base_asset == buy_gem.symbol and quote_asset == pay_gem.symbol:
                return buy_gem, pay_gem

        return None, None
